#
# Copyright (c) 2020 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.tools.gxx import gxx
from faber.action import CallError
from faber.utils import capture_output
from test.compiler import make_source, check_makedep, check_compile
import pytest


@pytest.fixture
def cxx():
    if not gxx.instances():
        pytest.skip('no g++ compiler found')
    else:
        return gxx.instance()


def test_makedep(cxx):
    """Check dependency detection."""

    header = make_source('foo.h', '// nothing')  # noqa F841
    src = make_source('foo.cc', '#include "foo.h"')
    check_makedep(cxx, src, 1)


def test_makedep_error(cxx):
    """Check missing header error."""

    src = make_source('foo.cc', '#include "foo.h"')
    with capture_output() as (out, err), pytest.raises(CallError):
        check_makedep(cxx, src, 1)
    stdout = out.getvalue()
    stderr = err.getvalue()
    print(stdout)
    print(stderr)


def test_compile(cxx):
    """Check dependency detection."""

    src = make_source('foo.cc', 'int main() {}')
    check_compile(cxx, src)


def test_compile_error(cxx):
    """Check dependency detection."""

    src = make_source('foo.cc', 'int main() { error }')
    with capture_output() as (out, err), pytest.raises(CallError):
        check_compile(cxx, src)
    stdout = out.getvalue()
    stderr = err.getvalue()
    print(stdout)
    print(stderr)



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact
from ..rule import rule
from ..tools.archiver import archiver
from ..artefacts.install import installed, installation as I
from ..artefacts.install import prefix, stage
from . import info
from .manifest import manifest
from os.path import join


class archive(artefact):

    def __init__(self, pkg, installation=None, format=None):
        """Create an archive.
        If 'installation' is None, make a source archive,
        otherwise a binary archive."""

        meta = info(pkg)
        name = '{}-{}'.format(meta.doc.name, meta.doc.version)
        self.format = format
        artefact.__init__(self, name + '.src' if not installation else name)
        s = join(self.module.builddir, 'packaging', self.id)
        if installation:
            # clone the installation into our own staging area
            i = installation(stage(s))
        else:
            features = (prefix(name), stage(s))
            m = manifest(self.module.srcdir, meta.doc.source)
            i = I('i:sinst',
                  [installed(f, features=features) for f in m],
                  features=features)
        rule(archiver.archive(format), self, i)

    @property
    def _filename(self):
        return self.name + archiver.extension(self.format)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from .feature import lazy_set
from .artefact import artefact
from .tool import tool
from .utils import add_metaclass
from .error import error_reporter
from os.path import join, exists
from os import makedirs
import logging

logger = logging.getLogger(__name__)


class proxy(object):
    """A module as seen when cross-referenced."""
    def __init__(self, module):
        self.__dict__.update(module._env)


class module_type(type):
    def __call__(cls, name, srcdir=None, builddir=None, process=True, **kwds):
        """Make sure to construct (and process) modules only once."""

        if name.startswith('.'):  # relative import
            if srcdir or builddir:
                raise ValueError('invalid arguments for relative imports')
            name, base = name[1:], module.current
            while name.startswith('.') and base:
                name, base = name[1:], base._parent
            if name.startswith('.'):
                raise ValueError('{} has no parent module'.format(base.name))
            name = '.'.join([base.name, name]) if base.name and name else \
                   name if name else base.name
            return proxy(cls._instances[name])
        elif name in cls._instances:
            return proxy(cls._instances[name])
        else:
            m = super(module_type, cls).__call__(name, srcdir, builddir, process, **kwds)
            return m


@add_metaclass(module_type)
class module(object):

    _instances = {}
    current = None

    @staticmethod
    def init(options, params):
        module.options = options
        module.params = params

    @staticmethod
    def finish():
        tool.finish()
        artefact.finish()
        module._instances.clear()

    def __init__(self, name, srcdir=None, builddir=None, process=True, **kwds):
        """Create a new module."""
        self.name = name
        if 'features' in kwds:
            self._features = kwds.pop('features').copy()
        else:
            self._features = lazy_set(module.params.copy())
        from . import builtin
        self._env = builtin.__dict__.copy()
        self._env['options'] = module.options
        self._env['features'] = self._features
        self._env.update(kwds)
        self._parent = module.current
        if self._parent:
            self.srcdir = join(self._parent.srcdir, srcdir or name)
            self.builddir = join(self._parent.builddir, builddir or name)
        else:
            self.srcdir = srcdir or name
            self.builddir = builddir or name
        module._instances[name] = self
        if process:
            with self:
                self.process()

    def __enter__(self):
        if self._parent:
            if self._parent._env['__name__']:
                self._env['__name__'] = self._parent._env['__name__'] + '.' + self.name
            else:
                self._env['__name__'] = self.name
        else:
            self._env['__name__'] = self.name
        self._env['srcdir'] = self.srcdir
        self._env['builddir'] = self.builddir
        module.current = self
        return self

    def __exit__(self, type, value, traceback):
        module.current = self._parent

    @property
    def features(self):
        return self._features

    def qname(self, a):
        """Return the given name, prefixed by this module's name"""
        name = self._env['__name__']
        return '{}.{}'.format(name, a) if name else a

    def process(self):
        script = join(self.srcdir, 'fabscript')
        if not exists(script):
            raise ValueError('{}: no such file'.format(script))
        # Run the script in env
        with error_reporter(script), open(script) as f:
            exec(f.read(), self._env)
        self._env.setdefault('default', [])
        self.__dict__.update({k: v for k, v in self._env.items() if not k.startswith('_')})
        self.default = self._env['default']

        # make sure the builddir exists
        if not exists(self.builddir):
            makedirs(self.builddir)

        return self.default



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import faber
from . import project
from . import logging
import os.path
import argparse


class PosArgsParser(argparse.Action):
    def __init__(self, option_strings, dest, **kwds):
        argparse.Action.__init__(self, option_strings, dest, **kwds)
    # Separates positional arguments into targets and parameters
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.parameters is None:
            namespace.parameters={}
        if not namespace.goals:
            namespace.goals=[]
        for v in values:
            if '=' in v:
                name, value = v.split('=', 1)
                if name in namespace.parameters:
                    if type(namespace.parameters[name]) is not list:
                        namespace.parameters[name] = [namespace.parameters[name]]
                    namespace.parameters[name].append(value)
                else:
                    namespace.parameters[name] = value
            else:
                namespace.goals.append(v)


def make_parser():

    parser = argparse.ArgumentParser(description='Faber is a construction tool.')
    parser.add_argument('goals', metavar='GOAL', nargs='*',
                        action=PosArgsParser,
                        help='a goal to update')
    parser.add_argument('parameters', metavar='PARAMETER=VALUE', nargs='*',
                        action=PosArgsParser,
                        help='a parameter value')
    parser.add_argument('--srcdir', default='.',
                        help='the location of the source directory')
    parser.add_argument('--builddir', default='.',
                        help='the location of the build directory')
    parser.add_argument('--rc', default=None,
                        help='the location of the rc file')
    log_args = parser.add_mutually_exclusive_group()
    log_args.add_argument('--log', action='append',
                          choices=logging.topics.keys(),
                          help='add log topic')
    log_args.add_argument('-l', dest='loglevel', type=int,
                          help='set log level (summary=1, actions=2, commands=4)')
    log_args.add_argument('-s', '--silent', action='store_true',
                          help='suppress all output')
    parser.add_argument('-j', '--parallel', type=int, default=1,
                        help='set concurrency level')
    parser.add_argument('--debug', action='store_true',
                        help='do not suppress traceback on error')
    parser.add_argument('-f', '--force', action='store_true',
                        help='update goals even if they are current')
    parser.add_argument('-n', '--noexec', action='store_true',
                        help='do not actually execute actions')
    parser.add_argument('-i', '--intermediates', action='store_true',
                        help='do not remove intermediate files after build')
    parser.add_argument('-c', '--clean', action='count',
                        help='cleanup')
    parser.add_argument('-t', '--timeout', type=int, default=0,
                        help='set timeout for individual actions')
    parser.add_argument('-p', '--profile', action='store_true',
                        help='log timing info with command')
    parser.add_argument('--info', choices=['goals', 'tools'], nargs='?', metavar='WHAT', const='goals',
                        help='print information about the build logic')
    parser.add_argument('--shell', action='store_true',
                        help='run interactive shell')
    parser.add_argument('-v', '--version', action='version', version=faber.__version__)
    return parser


def main():

    parser = make_parser()
    try:
        args, unknown = parser.parse_known_args()
        if args.silent:
            args.log = []

        def with_gen(w):
            k, v = w.split('=', 1) if '=' in w else (w, None)
            yield k[2:]
            yield v
        args.options = dict(map(with_gen, [w for w in unknown
                                           if w.startswith('--with-')]))
        args.options.update(dict.fromkeys([w[2:] for w in unknown
                                           if w.startswith('--without-')], None))

        unknown = [a for a in unknown
                   if not a.startswith('--with-') and not a.startswith('--without-')]
        # Parse the remainder, as there may be more positional arguments.
        # (See https://bugs.python.org/issue14191 for context)
        args = parser.parse_args(unknown, args)

        logging.setup(args.log, args.loglevel, args.debug, args.profile)
        if args.debug:
            faber.debug = True
        project.init()
        if args.rc:
            project.config(os.path.expanduser(args.rc))
        elif os.path.exists(os.path.expanduser('~/.faber')):
            project.config(os.path.expanduser('~/.faber'))

        if args.srcdir and os.path.exists(os.path.join(args.srcdir, '.faberrc')):
            project.config(os.path.join(args.srcdir, '.faberrc'))
        info = project.buildinfo(args.builddir, args.srcdir)
        if args.parameters and info.parameters and \
           args.parameters != info.parameters and \
           input('override existing parameters ? [y/N]:') != 'y':
            return False
        if args.options and info.options and \
           args.options != info.options and \
           input('override existing options ? [y/N]:') != 'y':
            return False
        info.parameters = args.parameters
        info.options = args.options
        proj = project.project(info,
                               parallel=args.parallel, force=args.force,
                               intermediates=args.intermediates,
                               timeout=args.timeout,
                               noexec=args.noexec)
        if args.info:
            result = proj.info(args.info, args.goals)
        elif args.shell:
            result = proj.shell()
        elif args.clean:
            result = proj.clean(args.clean)
        else:
            result = proj.build(args.goals)
        return result
    except KeyboardInterrupt:
        pass
    except Exception as e:
        if args.debug:
            raise
        else:
            print('Error: {}'.format(e))
        return False
    return True


def cli_main():
    """Convert boolean result to process exit status."""
    return 0 if main() else 1


if __name__ == "__main__":
    main()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from . import compiler
from ..feature import set
from ..action import action
import logging
import sys

logger = logging.getLogger('tools')


class cc(compiler.compiler):
    """C compiler base-class.
    As an abstract base-class it declares the actions all subclasses need to provide, without implementing them.

    Build scripts thus can reference `cxx.compile` et al., which the runtime will substitute by an appropriate
    compiler instance, if available (or fail to build)."""

    # Scan source files for header dependencies
    makedep = action()
    # Build object files from C source files.
    compile = action()
    # Build (static) library archives from object files.
    archive = action()
    # Link binaries (executables or shared libraries).
    link = action()

    @classmethod
    def instances(cls, fs=None):
        """Return all known C compiler instances for the current platform."""
        if cls is cc:
            if sys.platform == 'win32':
                from .msvc import msvc
                msvc.instances(fs)
        return super(cc, cls).instances(fs)

    @classmethod
    def instance(cls, fs=None):
        """Try to find a compiler instance for the current platform."""

        fs = set.instantiate(fs)
        if cls is cc and not cc.instantiated(fs):
            # we can't instantiate this class directly, so try to find
            # a subclass...
            logger.info('trying to instantiate a default C compiler')
            if sys.platform == 'win32':
                cc.try_instantiate('msvc', fs)
            cc.try_instantiate('gcc', fs)
            cc.try_instantiate('clang', fs)
            if not cc.instantiated(fs):
                msg = 'no C compiler found'
                msg += ' matching {}.'.format(fs.essentials()) if fs else '.'
                raise RuntimeError(msg)
        return super(cc, cls).instance(fs)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from .asyncio import *  # noqa F401
__backend__ = 'asyncio'



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from . import scheduler
from .feature import lazy_set
from .feature.condition import expr as fexpr
from .artefact import artefact, init as artefact_init
from .assembly import init as assembly_init
from .module import module
from .error import error_reporter
from .utils import aslist
from . import config as C
import logging
from configparser import ConfigParser
import warnings
from os.path import exists, join, abspath
import shutil
import os

builtin_formatwarning = warnings.formatwarning


def formatwarning(message, category, *args, **kwds):
    if category is UserWarning:
        # ignore everything except the message
        return 'Warning: {}\n'.format(message)
    else:
        return builtin_formatwarning(message, category, *args, **kwds)


warnings.formatwarning = formatwarning
logger = logging.getLogger(__name__)


def init():
    assembly_init()


def config(script):
    """Read in a config file and return its content."""

    logger.debug('reading config file {}'.format(script))
    if not exists(script):
        raise RuntimeError('config file "{}" not found'.format(script))
    env = {}
    with error_reporter(script), open(script) as f:
        exec(f.read(), env)
    return env


class buildinfo(object):
    """The collection of build-related information, initially provided
    via command-line arguments, stored in the build directory and later
    retrieved from there."""

    def __init__(self, builddir, srcdir=None):
        filename = join(builddir, '.faber', 'info') if builddir else None
        if filename and exists(filename):  # this is an existing build directory
            self.builddir = builddir
            c = ConfigParser(allow_no_value=True)
            c.read(filename)
            self.srcdir = c['general']['srcdir']
            if srcdir and abspath(srcdir) != abspath(self.srcdir):
                raise Exception('incompatible source directory')
            self.options = dict(c.items('options'))
            self.parameters = dict(c.items('parameters'))
        elif builddir and exists(join(builddir, 'fabscript')):  # it's a source directory
            self.builddir = builddir
            if srcdir and abspath(self.builddir) != abspath(srcdir):
                raise Exception('incompatible source directory')
            else:
                self.srcdir = srcdir or self.builddir
                self.options = {}
                self.parameters = {}
        else:  # it's a new builddir
            self.builddir = builddir
            self.srcdir = srcdir
            self.options = {}
            self.parameters = {}

    def store(self):
        if not exists(join(self.builddir, '.faber')):
            os.makedirs(join(self.builddir, '.faber'))
        filename = join(self.builddir, '.faber', 'info')
        config = ConfigParser(allow_no_value=True)
        config['general'] = dict(srcdir=self.srcdir)
        config['options'] = self.options
        config['parameters'] = self.parameters
        with open(filename, 'w') as info:
            config.write(info)

    def valid(self): return self.srcdir is not None


class options(dict):
    def __readonly__(self, *args, **kwargs):
        raise RuntimeError('options is immutable')
    __setitem__ = __readonly__
    __delitem__ = __readonly__
    pop = __readonly__
    popitem = __readonly__
    clear = __readonly__
    update = __readonly__
    setdefault = __readonly__
    del __readonly__

    def get_with(self, value):
        return self.get('with-' + value)

    def get_without(self, value):
        return self.get('without-' + value)


class project(object):

    def __init__(self, info, **kwds):
        """Construct a project. Parameters:

        * srcdir: the source directory
        * builddir: the build directory
        * options: any config options being passed (--with=, --without=)
        * parameters: dictionary of pre-defined parameters"""

        self.srcdir = info.srcdir
        self.builddir = info.builddir
        self.options = options(info.options)
        self.parameters = dict(info.parameters.items())
        info.store()
        self.sched_opts = kwds

    def __enter__(self):
        artefact_init()
        scheduler.init(self.parameters, self.builddir, **self.sched_opts)
        module.init(self.options, self.parameters)
        C.init(self.builddir)
        return self

    def __exit__(self, type, value, traceback):
        C.finish()
        module.finish()
        scheduler.finish()

    def build(self, goals):
        """build the project, updating the given goals or any defaults if None."""

        with self:
            m = module('', self.srcdir, self.builddir)
            if goals:
                try:
                    goals = [a for g in goals for a in artefact.lookup(g)]
                except KeyError as e:
                    print('don\'t know how to make {}'.format(e))
                    goals = []
                    result = False
            else:
                goals = aslist(m.default)
                # if we pick up default goals, check their conditions first
                deps = set([d for g in goals for d in g.features.dependencies()])
                # (allow dependencies to fail)
                scheduler.update(list(deps))

                def check(a):
                    if a.condition is None:
                        return True
                    elif isinstance(a.condition, fexpr):
                        return a.condition(a.features.eval())
                    else:
                        return a.condition

                # now filter by condition
                goals = [g for g in goals if check(g)]
                result = True
            if goals:
                result = scheduler.update(goals)
            elif result:
                print('no goals given and no "default" artefact defined - nothing to do.')
                result = True
        return result

    def clean(self, level):
        """Clean up file artefacts."""

        with self:
            m = module('', self.srcdir, self.builddir)  # noqa F841
            scheduler.clean(level)
            C.clean(level)
        if level > 1:
            shutil.rmtree(join(self.builddir, '.faber'))
        return True

    def info(self, what, items):
        """print project information, rather than performing a build.
        Parameters are the same as for the `build` function."""

        with self:
            result = True
            if what == 'goals':
                m = module('', self.srcdir, self.builddir)
                print('known artefacts:')
                for a in sorted(artefact.iter(), key=lambda a: a.qname):
                    print('  {}'.format(a.qname))
                if items:
                    try:
                        goals = [a for i in items for a in artefact.lookup(i)]
                    except KeyError as e:
                        print('don\'t know how to make {}'.format(e))
                        goals = []
                        result = False
                else:
                    goals = aslist(m.default)
                if goals:
                    scheduler.print_dependency_graph(goals)
            elif what == 'tools':
                from . import tool
                features = lazy_set(module.params.copy())
                for i in items:
                    tool.info(i, features)
        return result

    def shell(self):
        """load module in srcdir but rather than updating any goals,
        simply drop into an interactive shell."""

        import pydoc

        class Helper(pydoc.Helper):
            def __init__(self, env):
                self._env = env
                super(Helper, self).__init__()

            def intro(self):
                self.output.write("""
Welcome the Faber shell's help utility!

To get a list of available artefacts or features, type
"artefacts" or "features".""")

            def help(self, request):
                if isinstance(request, str):
                    request = request.strip()
                    if request == 'artefacts': self.listartefacts()
                    elif request == 'features': self.listfeatures()
                    else:
                        return super(Helper, self).help(request)
                else:
                    return super(Helper, self).help(request)
                self.output.write('\n')

            def listartefacts(self):
                self.list([k for k, v in self._env.items()
                           if isinstance(v, artefact)])

            def listfeatures(self):
                from .feature import feature
                self.list(feature._registry.keys())

        with self:
            m = module('', self.srcdir, self.builddir)
            # Use the module's environment, but inject a few helper functions
            env = m._env.copy()
            env['artefacts'] = list(artefact.iter())
            pydoc.help = Helper(env)
            history = None
            import code
            try:
                import readline
            except ImportError:
                pass
            else:
                import rlcompleter
                readline.set_completer(rlcompleter.Completer(env).complete)
                readline.parse_and_bind('tab: complete')
                fdir = join(self.builddir, '.faber')
                if not exists(fdir):
                    os.makedirs(fdir)
                history = join(fdir, 'history')
                try:
                    readline.read_history_file(history)
                except IOError:
                    pass

            code.interact('Faber interactive shell', local=env)
            if history:
                readline.write_history_file(history)
        return True



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber import scheduler
import contextlib
import tempfile
import shutil
import os
from os.path import join
import sys


def pyecho(targets, sources):
    vars = scheduler.variables(targets[0])
    vars = ' '.join(['{}={}'.format(n, v) for n, v in vars.items()])
    print('{} <- {}'.format(' '.join([t.name for t in targets]),
                            ' '.join([s.name for s in sources])) +
          (' ({})'.format(vars) if vars else ''))


@contextlib.contextmanager
def tempdir():
    dirpath = tempfile.mkdtemp()
    try: yield dirpath
    finally: shutil.rmtree(dirpath)


@contextlib.contextmanager
def cwd(tmp):
    # temporarily change into a working directory
    orig= os.getcwd()
    os.chdir(tmp)
    try: yield
    finally: os.chdir(orig)


def write_fabscript(dirpath, content):
    with open(join(dirpath, 'fabscript'), 'w') as f:
        f.write(content)


@contextlib.contextmanager
def argv(args):

    orig = sys.argv
    sys.argv = type(orig)(args)
    try: yield
    finally: sys.argv = orig



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..utils import capture_output
from .artefact import dependency_error
import asyncio
from concurrent.futures import TimeoutError
import subprocess
import tempfile
import sys
import logging
import os
import re
import locale

logger = logging.getLogger('scheduler')
encoding = locale.getpreferredencoding(False)


def quote(token):
    return '"{}"'.format(token)


def command_string(func, targets, sources, kwds):
    """Make a string of the command to be executed,
    for reporting purposes."""
    args= [repr(targets[0])] if len(targets) == 1 \
        else [] if not targets else [repr(targets)]
    if sources:
        args.append(repr(sources[0]) if len(sources) == 1
                    else repr(sources))
    if kwds:
        args.append(', '.join(['{}={}'.format(k, repr(v))
                               for k, v in kwds.items()]))
    return '{}({})'.format(func.__name__, ', '.join(args))


class recipe(object):

    @classmethod
    def init(cls, jobs=1, timeout=0, noexec=False):
        cls.semaphore = asyncio.BoundedSemaphore(jobs)
        cls.timeout = timeout or None
        cls.noexec = noexec

    @classmethod
    def finish(cls): pass

    def __init__(self, action, targets, sources):
        self.action = action
        self.targets = targets
        self.sources = sources
        self.status = None
        self.stdout = None
        self.stderr = None

    async def __call__(self):

        if callable(self.action.command):
            self.status, self.stdout, self.stderr = self.run_callable()
        else:
            self.status, self.stdout, self.stderr = await self.run_async_subprocess()
        return self.status

    def variables(self):
        return {k: [v] for k, v in self.action.map(self.targets[0].frontend.features).items()}

    def run_callable(self):

        from ..action import CallError
        # Setting '__noexec__' allows a function to be run even in noexec mode.
        if recipe.noexec and not hasattr(self.action.command, '__noexec__'):
            return True, '', ''
        targets = [t.frontend for t in self.targets]
        sources = [s.frontend for s in self.sources]
        vars = self.variables()
        cmd = command_string(self.action.command, targets, sources, vars)
        with capture_output() as (out, err):
            status = True
            try:
                status = self.action.command(targets, sources)
                # let users indicate failure by explicitly returning 'False'
                if status is not False:
                    status = True
            except dependency_error:
                # dependency errors are fatal - there is no point
                # in carrying on...
                raise
            except CallError as e:
                status = False
                cmd = e.cmd
            except Exception as e:
                # while normal exceptions simply result in update
                # failures.
                print(e)
                import traceback as tb
                tb.print_exc()
                status = False
        stdout = out.getvalue()
        stderr = err.getvalue()
        self.action.__status__([t.frontend for t in self.targets],
                               status, cmd, 0, stdout, stderr)
        return status, stdout, stderr

    async def run_async_subprocess(self):
        vars = self.variables()
        async with recipe.semaphore:
            cmd = self.action.command
            # substitute $(<[N])
            for m in re.findall(r'(\$\(<\[(\d+)\]\))', cmd):
                cmd = cmd.replace(m[0], quote(self.targets[int(m[1])].boundname))
            # substitute $(>[N])
            for m in re.findall(r'(\$\(>\[(\d+)\]\))', cmd):
                cmd = cmd.replace(m[0], quote(self.sources[int(m[1])].boundname))

            vars.update([('<', [quote(t.boundname) for t in self.targets])])
            vars.update([('>', [quote(s.boundname) for s in self.sources])])
            for v in vars:
                cmd = cmd.replace('$({})'.format(v), ' '.join(vars.get(v, [])))
            # cmd.exe can't deal with multi-line commands, so use a temporary bat file.
            bat = None
            if sys.platform == 'win32' and ('\n' in cmd or '\r' in cmd):
                bat = tempfile.NamedTemporaryFile(suffix='.bat', mode='w', delete=False)
                with bat:
                    bat.write(cmd)
                process = await asyncio.create_subprocess_exec(f'{bat.name}',
                                                               stdout=asyncio.subprocess.PIPE,
                                                               stderr=asyncio.subprocess.PIPE)
            else:
                process = await asyncio.create_subprocess_shell(cmd,
                                                                stdout=asyncio.subprocess.PIPE,
                                                                stderr=asyncio.subprocess.PIPE)
            try:
                stdout, stderr = await asyncio.wait_for(process.communicate(), timeout=recipe.timeout)
                stdout = stdout and stdout.decode(encoding).strip()
                stderr = stderr and stderr.decode(encoding).strip()
                status = process.returncode == 0
            except TimeoutError:
                status, stdout, stderr = False, '', ''
            finally:
                if bat:
                    os.unlink(bat.name)
            self.action.__status__([t.frontend for t in self.targets],
                                   status, cmd, 0, stdout, stderr)
        return status, stdout, stderr

    @staticmethod
    def run_subprocess(cmd):
        # cmd.exe can't deal with multi-line commands, so use a temporary bat file.
        bat = None
        if sys.platform == 'win32' and ('\n' in cmd or '\r' in cmd):
            bat = tempfile.NamedTemporaryFile(suffix='.bat', mode='w', delete=False)
            with bat:
                bat.write(cmd)
            process = subprocess.Popen(['cmd.exe', '/Q', '/C', bat.name],
                                       shell=False,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
        else:
            process = subprocess.Popen(cmd,
                                       shell=True,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        stdout = stdout and stdout.decode(encoding).strip()
        stderr = stderr and stderr.decode(encoding).strip()
        status = process.returncode == 0
        if bat:
            os.unlink(bat.name)
        return status, stdout, stderr



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import sys
import inspect
from contextlib import contextmanager

_script = None


class ScriptError(ValueError):

    def __init__(self, msg, location=None, level=0):
        if not location:
            f, filename, lineno, f, c, i = inspect.stack()[level]
            if filename == '<string>':
                filename = _script
            location = filename, lineno
        self.filename, self.lineno = location
        self.message = msg
        ValueError.__init__(self, '{}:{}: {}'.format(self.filename, self.lineno, msg))


class ArgumentError(ValueError): pass


@contextmanager
def error_reporter(script):
    global _script
    from . import debug

    _script = script
    try:
        yield
    except SyntaxError as e:
        if debug:
            raise
        filename, lineno = e.filename, e.lineno
        offset = e.offset or 1  # sometimes e.offset is None...
        value = 'invalid syntax:\n{}{}\n{}'.format(e.text, (offset - 1) * ' ' + '^', e.msg)
        if filename == '<string>':
            filename = script
        raise ScriptError(value, location=(filename, lineno))
    except (ScriptError, SystemExit, KeyboardInterrupt):
        raise  # pass through
    except Exception:
        if debug:
            raise
        type_, value, tb = sys.exc_info()
        f, filename, lineno, f, c, i = inspect.getinnerframes(tb)[-1]
        if filename == '<string>':
            filename = script
        raise ScriptError(str(value), location=(filename, lineno))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..utils import add_metaclass
from . import incidental
from .feature import feature
from .value import value, composite_value
from .condition import value as expr
from ..delayed import delayed
from ..error import ScriptError
from operator import iadd, ior
from functools import reduce


class set_type(type):

    def __getattr__(cls, name):
        # Allow the construction of expressions to support conditional values
        return expr(name)


@add_metaclass(set_type)
class set(object):

    @staticmethod
    def instantiate(features):
        """Convert the argument into a set."""
        if isinstance(features, set):
            return features
        elif not features:
            return set()
        else:
            return set(features)

    def __init__(self, *args):

        # if only one argument is provided and it's not a feature-value,
        # assume it's a container of values.
        if len(args) == 1 and not isinstance(args[0], (value, delayed)):
            args = args[0]

        self._features = {}
        self._delayed = []
        self._conditionals = []
        reduce(lambda s, a: ior(s, a), args, self)

    def keys(self):
        return self._features.keys()

    def values(self):
        return self._features.values()

    def items(self):
        return self._features.items()

    def get(self, name, default):
        return self._features.get(name, default)

    def dependencies(self):
        return [d._artefact for d in self._delayed] + \
            [c._artefact for c in self._conditionals
             if hasattr(c, '_artefact')]  # yuck !!

    def copy(self):
        """Create a copy of this set."""
        return set(*list(self._features.values()) + self._delayed + self._conditionals)

    def essentials(self):
        """Create a copy of this set containing only the essential
        (i.e., non-incidental) features."""
        return set(*[v for v in self._features.values()
                     if isinstance(v, composite_value) or not v._type.attributes & incidental])

    def update(self, other):
        """Add new features from `other`, replacing existing ones."""
        if isinstance(other, delayed):
            self._delayed.append(other)
        elif isinstance(other, value):
            self._features[other._name] = other
        else:
            other = set.instantiate(other)
            # only ordinary values can be updated
            for k, v in other._features.items():
                self._features[k] = v.copy()
            self._delayed[:] = other._delayed
            self._conditionals[:] = other._conditionals
        return self

    def eval(self, update=True):
        if update:
            from .. import scheduler
            # (allow dependencies to fail)
            scheduler.update(self.dependencies())
        # compute any delayed values
        self |= set(*[d.result() for d in self._delayed])
        self._delayed[:] = []
        if not self._conditionals:
            return self
        # evaluate conditional features
        initial = self.copy()
        current = initial.copy()
        for i in range(len(self._conditionals) + 1):
            added = [f for c in self._conditionals for f in c.eval(current)]
            fs = initial + set(*added)
            if current._features == fs._features:
                self |= set(*added)
                return self  # we are done !
            else:
                current = fs
        # didn't converge, conditions can't be met
        raise ScriptError('cannot satisfy conditionals')  # TODO: list location of all conditionals

    def __getattr__(self, name):
        if name in self._features:
            return self._features[name]
        elif name.startswith('__'):  # no special handling of special attributes
            return None
        else:
            raise ScriptError('no feature "{}" in set'.format(name), level=2)

    def __delattr__(self, name):
        del self._features[name]

    def __getitem__(self, name):
        return self._features[name]

    def __contains__(self, name):
        return name in self._features

    def matches(self, other):
        assert isinstance(other, set)
        for k, v in [(k, v) for k, v in self.items()]:
            if k in other and not other[k].matches(v):
                return False
        return True

    def __iadd__(self, other):
        """Add either a feature or merge an entire feature-set."""
        return self._cassign(iadd, other)

    def __add__(self, other):
        """Add either a feature or merge an entire feature-set."""
        fs = self.copy()
        fs += other
        return fs

    def __ior__(self, other):
        """Add either a feature or merge an entire feature-set."""
        return self._cassign(ior, other)

    def __or__(self, other):
        """Add either a feature or merge an entire feature-set."""
        fs = self.copy()
        fs |= other
        return fs

    def _cassign(self, op, other):
        """compound-assign: use `op` to add `other` to `self`"""
        if isinstance(other, delayed):
            self._delayed.append(other)
        elif isinstance(other, value):
            if other._condition is not None:
                self._conditionals.append(other)
            else:
                if other._name not in self._features:
                    self._features[other._name] = other.copy()
                else:
                    op(self._features[other._name], other)
        elif isinstance(other, set):
            for k, v in other._features.items():
                if k not in self._features:
                    self._features[k] = v.copy()
                else:
                    op(self._features[k], v)
            self._delayed += other._delayed
            self._conditionals += other._conditionals
        else:
            raise ValueError('invalid type {}'.format(type(other)))
        return self

    def __iter__(self):
        return iter(self.items())

    def __len__(self):
        return len(self._features)

    def _serialize(self):
        """Helper method for pretty-printing sets."""

        return reduce(lambda d, o: d.update(o.serialize()) or d, self.values(), {})

    def __str__(self):
        cln = self.__class__.__name__
        return '<{} {}>'.format(cln, ' '.join(['{}={}'.format(k, v)
                                               for k, v in sorted(self._serialize().items())]))

    def __repr__(self):
        cln = self.__class__.__name__
        return '<{} {}>'.format(cln, self._features)


def def_lazy_set():
    """Construct a lazy-set class that adds parameters on-demand."""

    def __init__(self, params, *args):
        self._params = params.copy()
        set.__init__(self, *args)

    def copy(self):
        return type(self)(self._params, *list(self._features.values()) + self._delayed + self._conditionals)

    def _parse(self):
        """Parse a dictionary of (command-line) parameters into
        a set of feature values."""
        converted = []
        for k, v in self._params.items():
            c = k.split('.')
            root = c[0]
            if root not in feature._registry:
                continue  # not defined yet, skip
            else:
                converted.append(k)
            # get an instance of the root feature
            if root not in self._features:
                f = feature.lookup(root)()
                self._features[root] = f
            sf = f = self._features[root]
            # find the designated subfeature
            for i in range(1, len(c)):
                sf = getattr(sf, c[i])
            # and set the given value
            sf += sf._type(*v if type(v) is list else [v])
        # clean up
        for c in converted:
            del self._params[c]

    def decorator(f):
        def wrapper(self, *args, **kwds):
            _parse(self)
            return f(self, *args, **kwds)
        return wrapper

    dict = {}
    for a, v in set.__dict__.items():
        if a in ('__add__',
                 '__contains__',
                 '__getattr__',
                 '__getitem__',
                 '__iadd__',
                 '__ior__',
                 '__iter__',
                 '__keys__',
                 '__len__',
                 '__or__',
                 'eval',
                 'items',
                 'matches',
                 'values'):
            dict[a] = decorator(v)
    dict['__init__'] = __init__
    dict['copy'] = copy
    return set_type('lazy_set', (set,), dict)


lazy_set = def_lazy_set()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import, print_function
from faber.project import init, config, project, buildinfo
from faber.module import module as M
from faber import logging
from test.common import tempdir
from os.path import join, exists, expanduser
from os import mkdir
import pytest


def pytest_addoption(parser):
    parser.addoption('--faber-log', action='append',
                     choices=logging.topics.keys(),
                     help='add log topic')
    parser.addoption('--compiler', action='append', default=[],
                     help='specify a compiler to test with. Ex: "gcc", "msvc", "native"')


def pytest_configure(config):
    logging.setup(log=config.getoption('faber_log'))


def pytest_generate_tests(metafunc):
    if 'compiler' in metafunc.fixturenames:
        compilers = metafunc.config.option.compiler or ['']
        metafunc.parametrize('compiler', compilers)


@pytest.fixture(autouse=True)
def session():
    # reset global state
    init()
    if exists(expanduser('~/.faber')):
        config(expanduser('~/.faber'))


@pytest.fixture()
def module():
    with tempdir() as root:
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        builddir = join(root, 'test-build')
        info = buildinfo(builddir, srcdir)
        with project(info):
            with M('test', srcdir, builddir, process=False):
                yield



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.module import module
from test.common import tempdir, write_fabscript
import os
import os.path
import pytest


@pytest.mark.usefixtures('module')
def test_simple():

    script="""
from faber.tools.compiler import include
i = include('.')"""

    with tempdir() as dirpath:
        write_fabscript(dirpath, script)
        m = module(dirpath)
        assert dirpath in m.i


@pytest.mark.usefixtures('module')
def test_paths():

    outer="""
from faber.tools.compiler import include, linkpath
i = include('.')
l = linkpath('.', base=builddir)
inner = module('inner')"""

    inner="""
from faber.tools.compiler import include, linkpath
i = include('.')
l = linkpath('.', base=builddir)
outer = module('..')
"""

    srcdir = module.current.srcdir
    subdir = os.path.join(srcdir, 'inner')
    os.mkdir(subdir)
    write_fabscript(srcdir, outer)
    write_fabscript(subdir, inner)
    m = module(srcdir, builddir='build')
    assert m.i == [srcdir]
    assert m.l == [m.builddir]
    assert m.inner.i == [subdir]
    assert m.inner.l == [os.path.join(m.builddir, 'inner')]
    assert m.inner.outer.i == [srcdir]
    assert m.inner.outer.l == [m.builddir]


@pytest.mark.usefixtures('module')
def test_kwds():

    outer="""
from faber.tools.compiler import include, linkpath
value = 24
inner = module('inner', value=value)"""

    inner="""
"""

    with tempdir() as root:
        subdir = os.path.join(root, 'inner')
        os.mkdir(subdir)
        write_fabscript(root, outer)
        write_fabscript(subdir, inner)
        m = module(root, builddir='build')
        assert m.inner.value == m.value



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, source
from ..rule import rule, depend
from ..tools.compiler import compiler
from os.path import join, splitext


class scan(artefact):
    """Scan a source file for dependencies and inject them into
    the dependency graph."""

    def __init__(self, src, obj, recipe=None, features=(), module=None):

        src = source.instantiate(src, module=module)
        name = join(splitext(src.name)[0] + '.d')
        artefact.__init__(self, name, features=features, module=module,
                          logfile=obj.logfile)
        self._obj = obj
        if not recipe:
            c = compiler.check_instance_for_type(src.type, features)
            recipe = c.makedep
        rule(recipe, self, src)
        depend(self._obj, self)

    def reset(self):
        # do not reset this artefact as we can't undo its action
        pass

    def __status__(self, status):
        # the file format is simply a newline-separated list
        # of (header) filenames
        if status:
            headers = [h.strip() for h in open(self._filename).readlines()]
            depend(self._obj, [source.instantiate(h, module=self.module)
                               for h in headers])



#
# Copyright (c) 2019 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import source
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_unique_src():
    """Check that two sources associated with the same name actually
yield the same object."""

    with patch('faber.scheduler.define_artefact') as define:
        s1 = source('foo')
        s2 = source('foo')
        assert s1 is s2
    define.assert_called_once()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.feature import feature, incidental, map, join
from faber.action import action
from faber.artefact import artefact, notfile, always
from faber.tools import fileutils
from faber.rule import rule
from faber.utils import capture_output
from test.common import pyecho
from os.path import exists
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_call():
    a = action()
    b = artefact('b', attrs=notfile)
    c = artefact('c', attrs=notfile)
    with pytest.raises(ValueError) as e:
        a(b, c)
    assert 'not implemented' in str(e.value)

    with capture_output() as (out, err):
        a = action('echo', 'echo $(<)')
        a([b])
    assert out.getvalue().strip() == 'test.b'
    assert err.getvalue() == ''


@pytest.mark.usefixtures('module')
def test_call_index():
    """Check that commands can index target and source variables."""
    b = artefact('b', attrs=notfile)
    c = artefact('c', attrs=notfile)
    d = artefact('d', attrs=notfile)
    with capture_output() as (out, err):
        a = action('echo', 'echo $(<[1]) $(>[0])')
        a([b, c], [d])
    assert out.getvalue().strip() == 'test.c test.d'
    assert err.getvalue() == ''


@pytest.mark.usefixtures('module')
def test_recipe():
    """Check that an artefact's __recipe__ method is called to report
    the execution of the recipe updating it."""
    a = artefact('a', attrs=notfile|always)
    b = artefact('b', attrs=notfile|always)
    c = artefact('c', attrs=notfile|always)
    a = rule(pyecho, a)
    b = rule(pyecho, b, a)
    c = rule(pyecho, c, b)
    with patch('faber.action.action.__status__') as recipe:
        assert b.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'b <- a'
        assert c.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'c <- b'


@pytest.mark.usefixtures('module')
def test_variables():
    """Check that an action's variables are properly substituted."""

    variable = feature('variable', attributes=incidental)

    class A(action):

        var = map(variable, join)
        command = 'echo $(var)'

    a = artefact('a', attrs=notfile|always)
    b = artefact('b', attrs=notfile|always)
    c = artefact('c', attrs=notfile|always)
    echo = action('echo', 'echo $(variable)')
    pye = action('pyecho', pyecho, ['variable'])
    a = rule(A(), a, features=variable('A'))
    b = rule(echo, b, a, features=variable('B'))
    c = rule(pye, c, b, features=variable('C'))
    with patch('faber.action.action.__status__') as recipe:
        assert a.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'A'
        assert b.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'B'
        assert c.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == "c <- b (variable=['C'])"


@pytest.mark.usefixtures('module')
def test_compound():
    """Compound a command and a Python function into a single action."""

    class C(action):

        touch = fileutils.touch

        @staticmethod
        def command(targets, sources):
            f = targets[0]._filename
            if C.touch(targets, sources) and exists(f):
                with open(f, 'w') as out:
                    out.write('something')

    a = rule(C(), 'a')
    a.update()
    with open(a._filename, 'r') as f:
        assert f.readlines() == ['something']



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action, CallError
from ..feature import set as fset, map, translate, select_if
from ..artefact import artefact
from .. import types
from ..assembly import implicit_rule as irule
from ..utils import capture_output
from . import compiler
from .cc import cc
from .cxx import cxx, cxxstd
from ..artefacts.library import library
from os.path import basename, splitext, join, normpath, pathsep, exists, isabs, relpath
try:
    import winreg
except ImportError:  # python 2
    import _winreg as winreg
from collections import OrderedDict
from subprocess import *
from xml.dom.minidom import parseString
import logging
import os
import sys

logger = logging.getLogger('tools')


class makedep(action):

    # /showIncludes emits to stderr !
    command = 'cl /nologo $(cppflags) /showIncludes /EP $(>)'
    cppflags = map(compiler.cppflags)
    cppflags += map(compiler.define, translate, prefix='/D')
    cppflags += map(compiler.include, translate, prefix='/I"', suffix='"')


class makedep_wrapper(action):
    """This is a wrapper around `cl /showIncludes ...` to normalize the output and
    make it portable across compilers."""

    def __init__(self):
        self.cmd = makedep()
        action.__init__(self, self.cmd.name, self.makedep)

    def map(self, fs):
        return self.cmd.map(fs)  # just forward variables from makedep

    def makedep(self, targets, sources):
        with capture_output() as (out, err):
            try:
                status = self.cmd(targets, sources)
            except CallError:
                status = False
        stderr = err.getvalue()
        if status is False:
            # on error make sure to report stderr,
            # but only after stripping off the inclusion notes
            stderr = '\n'.join([l for l in stderr.split('\n')
                                if not l.startswith('Note: including file: ')])
            print(stderr, file=sys.stderr)
            return status
        # skip 'Note: including file: ', and remove dupliates
        headers = set([l[22:].lstrip() for l in stderr.split('\n')
                       if l.startswith('Note: including file: ')])
        # header paths are relative to the toplevel srcdir,
        # while we need them to be relative to the current module
        base = targets[0].module.srcdir
        rp = lambda f, base: f if isabs(f) else relpath(f, base)
        headers = [rp(h, base) + '\n' for h in headers]
        dfile = targets[0]._filename
        with open(dfile, 'w') as f:
            f.writelines(headers)


class compile(action):

    command = 'cl /nologo $(cppflags) $(cflags) $(cxxflags) /GR /MD /EHsc /c /Fo$(<) $(>)'
    cppflags = map(compiler.cppflags)
    cppflags += map(compiler.define, translate, prefix='/D')
    cppflags += map(compiler.include, translate, prefix='/I"', suffix='"')
    cflags = map(compiler.cflags)
    cxxflags = map(compiler.cxxflags)
    cxxflags += map(cxxstd, translate, prefix='/std:c++')


class link(action):

    command = 'link /nologo $(ldflags) /out:$(<) $(>) $(libs)'
    ldflags = map(compiler.ldflags)
    ldflags += map(compiler.linkpath, translate, prefix='/libpath:"', suffix='"')
    ldflags += map(compiler.link, select_if, 'shared', '/DLL')
    libs = map(compiler.libs, translate, suffix='.lib')

    def submit(self, targets, sources):
        # sources may contain object files as well as libraries
        # Separate the two, and add the libraries to the libs variable.

        src, linkpath, libs = msvc.split_libs(sources)
        linkpath = [compiler.linkpath(l, base='') for l in linkpath]
        libs = [compiler.libs(l) for l in libs]
        fs = fset(*libs + linkpath)
        for t in targets:
            t.features |= fs
        action.submit(self, targets, src)


class archive(action):

    command = 'lib /nologo /out:$(<) $(>)'


class msvc(cc, cxx):

    # available toolchains by version
    _toolchains = OrderedDict()
    known_archs = ['x86_64', 'x86']
    win_archs = {'x86_64': 'x64',
                 'x86': 'x86'}

    makedep = makedep_wrapper()
    compile = compile()
    archive = archive()
    link = link()

    @classmethod
    def split_libs(cls, sources):
        """split libraries from sources.
        Return (src, linkpath, libs)"""

        src = []
        libs = []
        linkpath = set()
        for s in sources:
            if isinstance(s, library):
                libs.append(s.filename.apply(lambda x: splitext(basename(x))[0]))
                linkpath.add(s.path)
            elif isinstance(s, artefact):
                src.append(s)
            else:
                raise ValueError('Unknown type of source {}'.format(s))
        return src, linkpath, libs

    def __init__(self, name='msvc', command=None, version='', features=()):

        features = fset.instantiate(features)
        if not version:
            version = self.find_version_requirement(features)
        if not version and len(msvc._toolchains):
            version = list(msvc._toolchains)[0]
        if version not in msvc._toolchains:
            raise ValueError(f'unknown MSVC version {version}')
        arch = str(features.target.arch) if 'target' in features else None
        if arch and arch not in msvc._toolchains[version]:
            raise ValueError(f'MSVC {version} does not support target architecture {arch}')
        if arch:
            product_dir, path = msvc._toolchains[version][arch]
            features |= compiler.target(os='Windows')
        else:
            arch, (product_dir, path) = list(msvc._toolchains[version].items())[0]
            features |= compiler.target(arch=arch, os='Windows')
        super(msvc, self).__init__(name=name, version=version)
        self.features |= features

        self.makedep.cmd.subst('cl', '"{}\\{}"'.format(path, 'cl'))
        self.compile.subst('cl', '"{}\\{}"'.format(path, 'cl'))
        self.archive.subst('lib', '"{}\\{}"'.format(path, 'lib'))
        self.link.subst('link', '"{}\\{}"'.format(path, 'link'))

        # Extract INCLUDE, LIB, and LIBPATH from setup script
        setup = join(product_dir, 'vcvarsall.bat')
        output = check_output([setup, msvc.win_archs[arch], '&', 'set']).decode()
        vars = dict(line.split('=', 1) for line in output.splitlines() if '=' in line)
        self.vars = {k: vars[k] for k in ('INCLUDE', 'LIB', 'LIBPATH')}
        include = compiler.include(*[i for i in self.vars['INCLUDE'].split(pathsep) if i])
        linkpath = compiler.linkpath(*[l for l in self.vars['LIB'].split(pathsep) if l])
        linkpath += compiler.linkpath(*[l for l in self.vars['LIBPATH'].split(pathsep) if l])
        self.features += include
        self.features += linkpath

        irule(self.compile, types.obj, types.c)
        irule(self.compile, types.obj, types.cxx)
        irule(self.archive, types.lib, types.obj)
        irule(self.link, types.bin, (types.obj, types.dso, types.lib))
        irule(self.link, types.dso, (types.obj, types.dso, types.lib))

    @classmethod
    def find_path(cls, product_dir, arch):
        setup = join(product_dir, 'vcvarsall.bat')
        output = check_output([setup, arch, '&', 'set']).decode()
        vars = dict(line.split('=', 1) for line in output.splitlines() if '=' in line)
        path = vars['Path']
        for p in path.split(pathsep):
            if exists(join(p, 'cl.exe')):
                return p

    @classmethod
    def discover(cls):

        # start with versions reported by `vswhere`
        root = os.environ['ProgramFiles(x86)']
        vswhere = join(root, 'Microsoft Visual Studio', 'Installer', 'vswhere.exe')
        try:
            vswhere_cmd = [vswhere,
                           '-products', '*',
                           '-requires', 'Microsoft.VisualStudio.Component.VC.Tools.x86.x64',
                           '-format', 'xml']
            logger.debug(f"executing {' '.join(vswhere_cmd)}")
            output = check_output(vswhere_cmd).decode()
            instances = parseString(output).getElementsByTagName('instance')
            for i in instances:
                installation_path = i.getElementsByTagName('installationPath')[0].childNodes[0].data
                version = i.getElementsByTagName('productDisplayVersion')[0].childNodes[0].data
                logger.debug(f'result: installation_path={installation_path}, version={version}')
                ipath = join(installation_path, 'VC', 'Auxiliary', 'Build')
                # This reports a different version, though the meaning of it isn't clear:
                # version = open(join(ipath, 'Microsoft.VCToolsVersion.default.txt')).read().strip()
                cls._toolchains[version] = OrderedDict()
                for arch in msvc.known_archs:
                    product_dir, path = ipath, None
                    try:
                        path = cls.find_path(ipath, msvc.win_archs[arch])
                    except Exception:
                        pass
                    if path:
                        cls._toolchains[version][arch] = (normpath(product_dir), normpath(path))
                        logger.info(f'vswhere discovered MSVC version={version} arch={arch}')
                if not cls._toolchains[version]:
                    # remove empty entries
                    del cls._toolchains[version]
        except Exception as e:
            logger.debug(f'{e}')
            pass

        # Known toolset versions, in order of preference.
        known_versions = ['15.0',
                          '14.0',
                          '12.0',
                          '11.0',
                          '10.0',
                          '10.0express',
                          '9.0',
                          '9.0express',
                          '8.0',
                          '8.0express',
                          '7.1',
                          '7.1toolkit',
                          '7.0',
                          '6.0']

        # Names of registry keys containing the Visual C++ installation path (relative
        # to "HKEY_LOCAL_MACHINE\SOFTWARE\\Microsoft").
        reg_keys = {'6.0': 'VisualStudio\\6.0\\Setup\\Microsoft Visual C++',
                    '7.0': 'VisualStudio\\7.0\\Setup\\VC',
                    '7.1': 'VisualStudio\\7.1\\Setup\\VC',
                    '8.0': 'VisualStudio\\8.0\\Setup\\VC',
                    '8.0express': 'VCExpress\\8.0\\Setup\\VC',
                    '9.0': 'VisualStudio\\9.0\\Setup\\VC',
                    '9.0express': 'VCExpress\\9.0\\Setup\\VC',
                    '10.0': 'VisualStudio\\10.0\\Setup\\VC',
                    '10.0express': 'VCExpress\\10.0\\Setup\\VC',
                    '11.0': 'VisualStudio\\11.0\\Setup\\VC',
                    '12.0': 'VisualStudio\\12.0\\Setup\\VC',
                    '14.0': 'VisualStudio\\14.0\\Setup\\VC',
                    '15.0': 'VisualStudio\\15.0\\Setup\\VC'}

        for version in known_versions:
            cls._toolchains[version] = OrderedDict()
            for arch in msvc.known_archs:
                product_dir, path = None, None
                for x64elt in ('', 'Wow6432Node\\'):
                    try:
                        with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                            'SOFTWARE\\{}Microsoft\\{}'
                                            .format(x64elt, reg_keys[version])) as key:
                            product_dir = winreg.QueryValueEx(key, "ProductDir")[0]
                            path = cls.find_path(product_dir, msvc.win_archs[arch])
                    except Exception:
                        pass
                if path:
                    cls._toolchains[version][arch] = (normpath(product_dir), normpath(path))
                    logger.info(f'discovered via registry MSVC version={version} arch={arch}')
            if not cls._toolchains[version]:
                # remove empty entries
                del cls._toolchains[version]

    @classmethod
    def instances(cls, fs=None):
        """Return all known MSVC instances."""

        if not msvc.instantiated():
            for v in msvc._toolchains:
                for a, _ in msvc._toolchains[v].items():
                    msvc(version=v, features=compiler.target(arch=a))
        return super(cc, cls).instances(fs)


# If this module is imported, assume we are running inside Windows
msvc.discover()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.tool import *
from faber.feature import *
import pytest


def test_tool():

    class A(tool):
        def __init__(self, name=None, features=()):
            tool.__init__(self, name=name, features=features)

    class B(A):
        pass

    class C(B):
        def __init__(self, name='CC', features=()):
            B.__init__(self, name=name, features=features)

    # just generate an A
    assert isinstance(A.instance(), A)
    # make sure A instances have a feature 'A' with A.name='A'
    assert isinstance(A.instance(set(A.tool(name='A'))), A)

    # test simple inheritance
    assert isinstance(B.instance(), A)
    assert isinstance(B.instance(), B)
    assert isinstance(B.instance(set(A.tool(name='B'))), B)
    assert isinstance(B.instance(set(B.tool(name='B'))), B)

    # test tools with explicit name
    assert isinstance(C.instance(), C)
    assert isinstance(C.instance(set(A.tool(name='CC'))), C)

    c1 = C('C1')  # noqa F841
    c2 = C('C2')  # noqa F841
    cc1 = C.instance(set(A.tool(name='C1')))
    cc2 = C.instance(set(A.tool(name='C2')))
    assert cc1.features.A.name == 'C1'
    assert cc2.features.A.name == 'C2'
    cc3 = C.instance(set(A.tool(name='CC')))  # noqa F841
    with pytest.raises(ValueError):
        cc = C.instance(set(A.tool(name='CCCC')))  # noqa F841


def test_action():

    class A(tool):

        a = action('some command')

    class B(A): pass

    assert A.a._cls is A
    assert B.a._cls is B
    assert A.a is not B.a

    a = A()
    b = B()

    assert a.a._cls is A
    assert a.a._tool is a
    assert b.a._cls is B
    assert b.a._tool is b



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import asyncio
import pytest


@pytest.mark.asyncio
async def test_timeout():
    """Test that timeout settings are respected, i.e. a command taking too long
    will be aborted and generate an appropriate error."""
    try:
        await asyncio.wait_for(asyncio.sleep(1), timeout=0.1)
    except asyncio.TimeoutError:
        pass



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from .utils import add_metaclass
from . import scheduler
from .feature import set, map
from .utils import aslist
from . import output
import re
import logging
import sys

action_logger = logging.getLogger('actions')
command_logger = logging.getLogger('commands')
output_logger = logging.getLogger('output')


def quote(token):
    return '"{}"'.format(token)


class CallError(Exception):
    """CallErrors may be used to pass the command string back
    up the call chain so it can be reported."""

    def __init__(self, cmd):
        Exception.__init__(self, 'Error calling {}'.format(cmd))
        self.cmd = cmd


class action_type(type):
    def __new__(cls, name, bases, dict):
        """Collect all maps in a private dict so we don't have to
        look them up each time we need them."""
        dict['_maps'] = {k: v for k, v in dict.items() if isinstance(v, map)}
        return super(action_type, cls).__new__(cls, name, bases, dict)


@add_metaclass(action_type)
class action(object):
    """An action is executed in order to (re-)generate an 'artefact'."""

    var_ex = re.compile(r'\$\((?P<variable>\w+)\)')

    @staticmethod
    def command_string(func, targets, sources, kwds):
        """Make a string of the command to be executed,
        for reporting purposes."""
        args= [repr(targets[0])] if len(targets) == 1 \
            else [] if not targets else [repr(targets)]
        if sources:
            args.append(repr(sources[0]) if len(sources) == 1
                        else repr(sources))
        if kwds:
            args.append(', '.join(['{}={}'.format(k, repr(v))
                                   for k, v in kwds.items()]))
        return '{}({})'.format(func.__name__, ', '.join(args))

    @property
    def qname(self):
        """The qualified name of this action."""
        if self._cls:
            return '{}.{}'.format(self._cls.__name__, self.name)
        else:
            return self.name

    @property
    def id(self):
        """Yield a unique identifier for this action."""
        return '%s-%x' % (self.qname, id(self))

    @property
    def tool(self):
        """The tool this action is a method of. (May be None)"""
        return self._tool

    @property
    def abstract(self):
        """The action is 'abstract' if it is a tool method
        without a tool instance."""
        return self._cls and not self._tool

    @property
    def features(self):
        return self._tool.features if self._tool else set()

    @property
    def path_spec(self):
        return self._tool.path_spec if self._tool else ''

    def __init__(self, *args):
        """Construct an action with one of these signatures:

          * action(command)
          * action(name, command)
          * action(name, command, vars)"""

        self.command = hasattr(self, 'command') and self.command
        self.name = self.__class__.__name__
        self.vars = []
        if len(args) == 0:
            pass
        elif len(args) == 1:
            self.command = args[0]
        elif len(args) == 2:
            self.name, self.command = args
        else:
            self.name, self.command, self.vars = args[:3]
        if type(self.command) is str and not self.vars:
            self.vars = action.var_ex.findall(self.command) if self.command else []
        self._cls = None
        self._tool = None

    def instantiate(self, features=()):
        """If this action is abstract, find a tool matching the given
        features and return the corresponding concrete action."""

        if self._cls and not self._tool:
            tool = self._cls.instance(features)
            return getattr(tool, self.name)
        else:
            return self

    def subst(self, old, new):
        self.command = self.command.replace(old, new, 1)

    def __call__(self, targets, sources=[], **kwds):
        """Explicitly call the action with the given artefacts."""

        # Make sure this action is instantiated.
        self = self.instantiate()

        tnames = [t.boundname for t in aslist(targets)]
        snames = [s.boundname for s in aslist(sources)]
        if not self.command:
            raise ValueError('{} not implemented'.format(self.qname))
        elif callable(self.command):
            status = self.command(targets, sources, **kwds)
            if status is False:
                cmd = action.command_string(self.command, tnames, snames, kwds)
                raise CallError(cmd)
            return status
        elif type(self.command) is str:
            cmd = self.command
            # substitute $(<[N])
            for m in re.findall(r'(\$\(<\[(\d+)\]\))', cmd):
                cmd = cmd.replace(m[0], quote(tnames[int(m[1])]))
            # substitute $(>[N])
            for m in re.findall(r'(\$\(>\[(\d+)\]\))', cmd):
                cmd = cmd.replace(m[0], quote(snames[int(m[1])]))
            # substitute $(<)
            cmd = cmd.replace('$(<)', ' '.join([quote(t) for t in tnames]))
            # substitute $(>)
            cmd = cmd.replace('$(>)', ' '.join(quote(s) for s in snames))
            if targets:
                vars = self.map(targets[0].features)
                for v in self.vars:
                    cmd = cmd.replace('$({})'.format(v), vars.get(v) or '')
            status, stdout, stderr = scheduler.run(cmd)
            if stdout:
                print(stdout)
            if stderr:
                print(stderr, file=sys.stderr)
            if not status:
                raise CallError(cmd)
            return status

    def submit(self, targets, sources):
        if not self.command:
            raise RuntimeError(None, ': {} is not implemented'.format(self.qname))
        scheduler.define_recipe(self, targets, sources)

    def map(self, fs):
        """translate the given feature-set using any map this action has defined."""
        fs.eval(update=False)
        if self._maps:
            return {k: v(fs) for k, v in self._maps.items()}
        else:
            return {v: str(fs[v]) if v in fs else '' for v in self.vars}

    def __status__(self, targets, status, command, time, stdout, stderr):
        """Report completion of the recipe."""
        logfile = targets[0].logfile
        targets = ' '.join([t.qname for t in targets])
        alog = output.coloured('{} {}'.format(self.qname, targets), attrs=['bold'])
        if logfile:
            logfile.write(alog + '\n')
            logfile.write(command + '\n')
            if stdout:
                logfile.write(stdout)
            if stderr:
                logfile.write(stderr)
        else:
            action_logger.info(alog)
            command_logger.info(command, extra={'time': time})
            if stdout:
                output_logger.info(stdout)
            if stderr:
                output_logger.info(stderr)



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.package import load_info
from test.common import tempdir
import pytest
from os.path import join


def test_info():

    info = """
package:
  name: the name
  version: 1.0
  description: some text here

a:
  - 1
  - 42
b: bla
c:
  answer: 42
  foo: bar
  key: value
"""

    with tempdir() as dirpath:
        filename = join(dirpath, 'test.pkg')
        with open(filename, 'w') as f:
            f.write(info)

        doc = load_info(filename)
        assert doc.package.name == 'the name'
        assert doc.package.version == 1.0
        assert doc.package.description == 'some text here'
        assert doc.a[0] == 1
        assert doc.c.answer == 42
        assert doc.c.key == 'value'


@pytest.mark.usefixtures('module')
def test_archive():
    pass



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, notfile, always
from .. import scheduler


class features(artefact):
    def __init__(self, name, features=()):
        artefact.__init__(self, name, attrs=notfile|always, features=features)

    def __status__(self, status):
        print(self.features.eval(update=False))


class dependency_graph(artefact):
    def __init__(self, name, a, dependencies=[], features=()):
        artefact.__init__(self, name, attrs=notfile|always, features=features)
        scheduler.declare_dependency(self, dependencies)
        self.a = a

    def __status__(self, status):
        scheduler.print_dependency_graph(self.a)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.test import test, report, pass_, fail, xpass, xfail
from faber.artefact import always, notfile
from faber.action import action
from faber.rule import rule
from faber import cli
from test.common import tempdir, write_fabscript, argv
import pytest


@pytest.mark.usefixtures('module')
def test_outcome():
    passing = rule(action('true', 'echo 1'), 'passing', attrs=always|notfile)
    failing = rule(action('false', 'unknown'), 'failing', attrs=always|notfile)
    test1 = test('test1', passing)
    test2 = test('test2', failing)
    test3 = test('test3', passing, expected=fail)
    test4 = test('test4', failing, expected=fail)
    assert not all([t.update() for t in [test1, test2, test3, test4]])
    assert test1.outcome == pass_
    assert test2.outcome == fail
    assert test3.outcome == xpass
    assert test4.outcome == xfail


@pytest.mark.usefixtures('module')
def test_report():
    passing = rule(action('true', 'echo 1'), 'passing', attrs=always|notfile)
    failing = rule(action('false', 'unknown'), 'failing', attrs=always|notfile)
    test1 = test('t1', passing)
    test2 = test('t2', failing)
    test3 = test('t3', passing, expected=fail)
    test4 = test('t4', failing, expected=fail)
    test5 = test('t5', failing, condition=False)
    r = report('report', [test1, test2, test3, test4, test5])
    assert r.update()
    assert test1.outcome == pass_
    assert test2.outcome == fail
    assert test3.outcome == xpass
    assert test4.outcome == xfail
    assert test5.outcome is None


@pytest.mark.parametrize('fail', [False, True])
def test_failure(fail):

    script="""
from faber.test import test, report
passing = rule(action('true', 'echo 1'), 'passing', attrs=always|notfile)
failing = rule(action('false', 'unknown'), 'failing', attrs=always|notfile)
test1 = test('t1', passing)
test2 = test('t2', failing)
r = report('report', [test1, test2], fail_on_failures={})
default = r
""".format(fail)

    with tempdir() as dirpath:
        write_fabscript(dirpath, script)
        command = ['faber']
        command.append('--srcdir={}'.format(dirpath))
        command.append('--builddir={}'.format(dirpath))
        with argv(command):
            if fail:
                assert not cli.main()
            else:
                assert cli.main()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..feature import feature, multi, path, incidental
from ..feature import set, map, join as fjoin
from ..tool import tool
from .compiler import include, ldflags, linkpath, libs, target, runpath
from .. import platform
from os.path import join
import subprocess
import re
import logging

logger = logging.getLogger('tools')

pythonpath = feature('pythonpath', attributes=multi|path|incidental)


class run(action):

    pythonpath = map(pythonpath, fjoin)
    runpath = map(runpath, fjoin)
    if platform.os == 'Windows':
        command = """set PATH=$(runpath);%PATH%
set PYTHONPATH=$(pythonpath)
python $(>)"""
    else:
        command = 'LD_LIBRARY_PATH=$(runpath) PYTHONPATH=$(pythonpath) python $(>)'


class python(tool):

    run = run()

    def check_python(self, cmd):
        return subprocess.check_output([self.command, '-c', cmd], universal_newlines=True).strip()

    def check_sysconfig(self, cmd):
        r = self.check_python(f'import sysconfig as c; print(c.{cmd})')
        return r if r != 'None' else ''

    def __init__(self, name='python', command=None, version='', features=()):
        if not isinstance(features, set):
            features = set(features)
        self.command = command or 'python'
        v = self.check_python('import platform; print(platform.python_version())')
        if version and v != version:
            raise ValueError(f'{self.command} version mismatch: expected {version}, got {v}')
        a = self.check_python('import platform; print(platform.machine())')
        b = int(self.check_python('import struct;print(struct.calcsize("P") * 8)'))
        if a == 'AMD64':  # Windows...
            a = 'x86_64'
        if a == 'x86_64' and b == 32:
            a = 'x86'
        if 'target' in features:
            arch = str(features.target.arch)
            if arch != a:
                raise ValueError(f'{self.command} architecture mismatch: expected {arch}, got {a}')
        else:
            features += target(arch=a)
        tool.__init__(self, name=name, version=v)
        self.features |= features
        self.run.subst('python ', self.command + ' ')
        # Now determine all the flags we may need to compile C / C++ extensions
        self.include = include(self.check_sysconfig('get_config_var("INCLUDEPY")'))
        platform = self.check_python('import platform; print(platform.system())')
        impl = self.check_python('import platform; print(platform.python_implementation())')
        prefix = self.check_python('import sys; print(sys.prefix)')
        if platform == 'Windows':
            if impl == 'CPython':
                version = self.check_sysconfig('get_config_var("py_version_nodot")')
                self.libfile = join(prefix, 'libs', f'python{version}.lib')
                self.libpath = join(prefix, 'libs')
                self.lib = 'python' + version
            elif impl == 'PyPy':
                version = self.check_python('import sys; print(sys.version_info[0])')
                self.libfile = join(prefix, 'libs', f'libpypy{version}-c.lib')
                self.libpath = join(prefix, 'libs')
                self.lib = 'libpypy' + version + '-c'
            else:
                raise ValueError('unsupported Python implementation')
        else:
            self.libpath = self.check_sysconfig('get_config_var("LIBDIR")')
            if impl == 'CPython':
                self.libfile = self.check_sysconfig('get_config_var("LIBRARY")')
                match = re.search(r'(python.*)\.(a|so|dylib)', self.libfile)
                if match:
                    self.lib = match.group(1)
                    if match.group(2) == 'a':
                        flags = self.check_sysconfig('get_config_var("LINKFORSHARED")')
                        if flags is not None:
                            flags=flags.split()
                            self.ldflags = ldflags(*flags)  # TODO: use them !
            elif impl == 'PyPy':
                version = self.check_python('import sys; print(sys.version_info[0])')
                suffix = self.check_sysconfig('get_config_var("SHLIB_SUFFIX")')
                self.libfile = join(prefix, 'libs', f'libpypy{version}-c{suffix}')
                self.libpath = join(prefix, 'libs')
                self.lib = 'pypy' + version + '-c'
            else:
                raise ValueError('unsupported Python implementation')

            # Only publish the libpath if it isn't a system path
            if self.libpath in ['/usr/lib', '/usr/lib64']:
                self.libpath = None
        self.linkpath = linkpath(self.libpath) if self.libpath else linkpath()
        # FIXME: figure out the exact conditions under which this needs to be defined
        # condition=(set.cc.name=='msvc')|(set.cxx.name=='msvc'))
        self.libs = libs(self.lib)
        flags = self.check_sysconfig('get_config_var("MODLIBS")')
        flags += ' ' + self.check_sysconfig('get_config_var("SHLIBS")')
        flags = [f[2:] for f in flags.strip().split() if f.startswith('-l')]
        self.libs += libs(*flags)
        version = self.check_sysconfig('get_config_var("py_version")')
        if version >= '3.5':
            self.ext_suffix = self.check_sysconfig('get_config_var("EXT_SUFFIX")')
        else:
            self.ext_suffix = self.check_sysconfig('get_config_var("SO")')



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from . import platform
from os.path import split, splitext, join
from collections import defaultdict
import logging

logger = logging.getLogger(__name__)


class typed_name(object):
    """adorn a (file-)name with a type"""

    def __init__(self, name, type):
        self.name = name
        self.type = type

    def __str__(self): return '<{} {}>'.format(self.type.name, self.name)


class type(object):

    # a dictionary mapping host names to extension->type mappings
    # As there are potentially multiple types using the same extension,
    # we only remember the first for reverse (extension->type) lookup.
    _register = defaultdict(dict)

    class maybemethod(object):
        """A descritor that allows a member to be accesses as either
        a method or a classmethod"""

        def __init__(self, f):
            self.f = f
        def __get__(self, obj, cls):
            return lambda *args: self.f(obj, *args)

    @staticmethod
    def discover(filename, host=''):
        """Discover a file's type from its file extension."""

        ext = splitext(filename)[1]
        ext = ext and ext[1:]
        if not ext: return None
        elif host in type._register and ext in type._register[host]:
            return type._register[host][ext]
        else:
            return type._register[''].get(ext)

    @maybemethod
    def typed_name(self, name, host=''):
        """Create a typed object. If this is called as a classmethod,
        look up the appropriate type from the file extension. Otherwise
        use `self` as type."""

        host = host or platform.os
        t = self or type.discover(name, host)
        if not t:
            raise RuntimeError('Cannot discover type for {} on {}'
                               .format(name, host))
        return typed_name(name, t)

    def synthesize_name(self, name, host=''):
        """Synthesize a new filename from an existing filename and the given type."""

        stem, ext = splitext(name)
        ext = ext and ext[1:]
        host = host or platform.os
        if self:
            t = self
        elif ext in type._register[host]:
            t = type._register[host][ext]
        else:
            t = type._register[''][ext]
        ext = t.ext(host)[0]
        return ext and stem + '.' + ext or stem

    def ext(self, host=''):

        return self._ext[host] if host in self._ext else self._ext['']

    def __init__(self, name, ext=None, **host_specific):
        """Define a (named) type with one or more filename extensions."""
        self.name = name
        self._ext = defaultdict(list)
        if ext is not None:
            host_specific[''] = ext
        for h in host_specific:
            for e in host_specific[h]:
                self._ext[h].append(e)
                if e not in type._register[h]:
                    type._register[h][e] = self
                    logger.debug('registering {} on {}'.format(name, h))
                else:
                    logger.debug('not registering {} on {}'.format(name, h))

    def __repr__(self):
        return '<type {}>'.format(self.name)


class library(type):
    """library names not only have a file extension, but may also be prefixed with 'lib'."""

    def synthesize_name(self, name, host=''):

        dir, base = split(name)
        host = host or platform.os
        base = type.synthesize_name(self, base, host)
        if host != 'Windows' or self.name == 'lib':
            # on Windows, only static libs get the 'lib' prefix
            base = 'lib' + base
        return dir and join(dir, base) or base


c = type('c', ['c'])
cxx = type('cxx', ['cc', 'cxx', 'cpp', 'C'])
obj = type('obj', ['o'], Windows=['obj'])
bin = type('bin', [''], Windows=['exe'])
lib = library('lib', ['a'], Windows=['lib'])
dso = library('dso', ['so'], Windows=['dll'], Darwin=['dylib'])
pyd = type('pyd', ['so'], Windows=['pyd'])



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..tool import tool
from ..feature import feature, multi, incidental, map


xsltflags = feature('xsltflags', attributes=multi|incidental)


class process(action):

    command = 'xsltproc $(xsltflags) -o $(<) $(stylesheet) $(>)'

    xsltflags = map(xsltflags)


class xsltproc(tool):

    process = process()

    def __init__(self, name='xsltproc', command=None, version='', features=()):
        tool.__init__(self, name=name, version=version)
        self.features |= features
        if command:
            self.process.subst('xsltproc', command)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact
from ..artefacts.library import library
from ..tool import tool
from ..feature import feature, multi, path, incidental
from .. import types
from ..error import ArgumentError
from os.path import basename
from importlib import import_module
import logging

logger = logging.getLogger('tools')

cppflags = feature('cppflags', attributes=multi|incidental)
define = feature('define', attributes=multi|incidental)
include = feature('include', attributes=multi|path|incidental)
cflags = feature('cflags', attributes=multi|incidental)
cxxflags = feature('cxxflags', attributes=multi|incidental)
ldflags = feature('ldflags', attributes=multi|incidental)
link = feature('link', ['static', 'shared'])
linkpath = feature('linkpath', attributes=multi|path|incidental)
libs = feature('libs', attributes=multi|incidental)
target = feature('target', feature(name='os', sub=True), feature(name='arch', sub=True))
runpath = feature('runpath', attributes=multi|path|incidental)
soname = feature('soname', attributes=incidental)


class compiler(tool):

    path_spec = '{compiler.name}-{compiler.version}/{target.arch}/{link}/'

    @classmethod
    def split_libs(cls, sources):
        """split libraries from sources.
        Return (src, linkpath, libs)"""
        src = []
        libs = []
        linkpath = set()
        for s in sources:
            if isinstance(s, library):
                libs.append(basename(s.libname))
                linkpath.add(s.path)
            elif isinstance(s, artefact):
                src.append(s)
            else:
                raise ValueError('Unknown type of source {}'.format(s))
        return src, linkpath, libs

    @staticmethod
    def check_instance_for_type(type, features=None):
        """Make sure we have a matching compiler for the given type."""
        name = {types.c: 'cc',
                types.cxx: 'cxx'}[type]
        mod = import_module('.{}'.format(name), 'faber.tools')
        return getattr(mod, name).instance(features)

    @classmethod
    def try_instantiate(cls, name, fs=None):
        """Try to instantiate the given compiler, but fail silently."""

        try:
            mod = import_module('.{}'.format(name), 'faber.tools')
            getattr(mod, name)(features=fs)
        except (SyntaxError, ArgumentError):  # these errors need to be reported.
            raise
        except Exception as e:
            logger.info('trying to instantiate {} yields "{}"'.format(name, e))



#
# Copyright (c) 2019 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ... import test
from ...artefact import notfile, always, internal
from ...rule import rule, depend
from ...action import action
from ...tools import compiler
from ...feature import map, join
from ... import platform
from ...utils import capture_output
import re


class test_module(test.suite):
    """A test module represents all the test cases contained
    in a single Boost.Test executable."""

    class run(action):

        runpath = map(compiler.runpath, join)
        if platform.os == 'Windows':
            command = """set PATH=$(runpath);%PATH%
$(>)"""
        else:
            command = 'LD_LIBRARY_PATH=$(runpath) $(>)'

        def __init__(self, args):
            super(test_module.run, self).__init__(test_module.run.command + ' ' + ' '.join(args))

    def __init__(self, name, exe, **kwds):
        super(test_module, self).__init__(name, [], **kwds)
        self.exe = exe
        a = rule(self.query, 't:' + self.name, sources=self.exe,
                 attrs=notfile|always|internal)
        depend(self, a)

    def reset(self):
        pass

    def query(self, target, source):

        runpath = 'linkpath' in target[0].features and compiler.runpath(str(target[0].features.linkpath))
        target[0].features += runpath
        a = self.run(['--list_content'])
        with capture_output() as (out, err):
            result = a(target, source)
        if result:
            stdout = out.getvalue()
            for name in self.parse(stdout):
                t = test.test(name, self.exe, run=self.run(['-x', 'no', '-t', name]),
                              module=self.module,
                              features=runpath)
                self.tests.append(t)
                depend(self, t)

    @staticmethod
    def parse(output):
        """Parse the output generated from boost.test using the --list_content option."""

        stack = []
        indents = [-1]
        for line in output.split('\n'):
            m = re.match(r'(?P<indent>\s*)(?P<name>[^*:\s]*)', line)
            indent, name = len(m.group(1)), m.group(2)
            if indent > indents[-1]:
                indents.append(indent)
                stack.append(name)
            elif indent == indents[-1]:
                yield '/'.join(stack)
                stack[-1] = name
            else:
                yield '/'.join(stack)
                while indent < indents[-1]:
                    indents.pop()
                    stack.pop()
                stack[-1] = name



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from os.path import join, normcase, basename
from os import walk
import fnmatch
from glob import glob


class manifest(set):

    def __init__(self, base, source):

        self.base = base
        # TODO: need to sort entries to handle excludes last
        for s in source:
            cmd = s.split(' ')
            args = map(normcase, cmd[1:])
            if cmd[0] == 'include':
                self._include(args)
            elif cmd[0] == 'recursive-include':
                self._recursive_include(args[0], args[1:])
            elif cmd[0] == 'exclude':
                self._exclude(args)
            elif cmd[0] == 'recursive-exclude':
                self._recursive_exclude(args[0], args[1:])
            else:
                raise ValueError('unrecognized command "{}"'
                                 .format(cmd[0]))

    def _include(self, patterns):
        for pattern in patterns:
            pattern = join(self.base, pattern)
            files = glob(pattern)
            self |= set(files)

    def _recursive_include(self, dir, patterns):
        base = join(self.base, dir)
        files = []
        for pattern in patterns:
            for root, _, filenames in walk(base):
                for filename in fnmatch.filter(filenames, pattern):
                    files.append(join(root, filename))
        self |= set(files)

    def _exclude(self, patterns):
        files = []
        for pattern in patterns:
            for filename in fnmatch.filter(self, pattern):
                files.append(filename)
        self -= set(files)

    def _recursive_exclude(self, dir, patterns):
        files = []
        if dir == '.':
            dir = ''
        for pattern in patterns:
            for filename in self:
                if (filename.startswith(dir) and
                    fnmatch.filter(basename(filename), pattern)):  # noqa E129
                    files.append(filename)
        self -= set(files)



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..feature import feature, path, incidental
from ..action import action
from ..tool import tool
import os
import os.path
import stat
import shutil

prefix = feature('prefix', attributes=path|incidental)
stage = feature('stage', attributes=path|incidental)


def copyfile(source, target):
    assert os.path.isfile(source)
    # copy content,...
    shutil.copyfile(source, target)
    # ...stat,...
    shutil.copystat(source, target)
    # ...owner, and group
    st = os.stat(source)
    if hasattr(os, 'chown'):  # not available on Windows
        os.chown(target, st[stat.ST_UID], st[stat.ST_GID])


def copydir(source, target):
    assert os.path.isdir(source)
    shutil.copytree(source, target, symlinks=True)


class install(action):

    def __init__(self, mode=None):
        action.__init__(self)
        self.mode = mode

    @staticmethod
    def command(targets, sources):
        t = targets[0]
        s = sources[0]
        if os.path.isfile(s._filename):
            copyfile(s._filename, t._filename)
        elif os.path.isdir(s._filename):
            copydir(s._filename, t._filename)
        else:
            raise ValueError('Cannot install "{}"; neither a file nor directory.'.format(s._filename))


def create_manifest(t, sources):
    stage = str(t[0].features.stage) if 'stage' in t[0].features else ''
    files = [s._filename[len(stage):] for s in sources
             if s._filename.startswith(stage)]
    with open(t[0]._filename, 'w') as manifest:
        for f in files:
            manifest.write(f + '\n')


class installer(tool):

    install_exe = install()
    install_data = install()
    install = install_data



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact
from ..rule import rule
from os.path import normpath, join, sep


class dir(artefact):

    def __init__(self, recipe, name, sources, *args, **kwds):
        dependencies = kwds.pop('dependencies', [])
        artefact.__init__(self, name, *args, **kwds)
        rule(recipe, self, sources, dependencies=dependencies)

    @property
    def _filename(self):
        # make sure this is recognized by tools (e.g. xsltproc) as a directory,
        # by explicitly appending a '/'
        return normpath(join(self.module.builddir, self.relpath, self.name)) + sep



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..feature import set, map, translate, select_if
from .. import types
from .. import platform
from ..assembly import implicit_rule as irule
from . import compiler
from .cc import *
import os.path
import os
import subprocess

# known architectures for each machine type
marchs = dict(x86_64=['x86_64', 'x86'],
              arm64=['arm64'],
              w64=['w64', 'w32'])

# compiler flags per architecture
arch_flags = dict(x86_64=['-m64'],
                  x86=['-m32'],
                  arm64=[],
                  w64=['-m64'],
                  w32=['-m32'])


def validate(cls, command, version, features):

    features = set.instantiate(features)
    version = version or cls.find_version_requirement(features)
    v = subprocess.check_output([command, '-dumpversion']).decode().strip()
    m = subprocess.check_output([command, '-dumpmachine']).decode().strip()
    if version and v != version:
        raise ValueError('{} version mismatch: expected {}, got {}'
                         .format(command, version, v))
    else:
        version = v
    cpu, _, os = platform.split_triplet(m)
    if os == 'mingw32':
        os = 'Windows'
        machine = 'w64'
    else:
        machine = cpu

    if machine not in marchs:
        raise ValueError('Unsupported machine type "{}"'.format(machine))
    if 'target' in features:
        arch = str(features.target.arch)
        if arch not in marchs[machine]:
            raise ValueError('Unsupported target architecture "{}"'.format(arch))
        features |= compiler.target(os=os)
    else:
        arch=marchs[machine][0]
        features |= compiler.target(arch=arch, os=os)

    features |= compiler.cxxflags(*arch_flags[arch])
    features |= compiler.ldflags(*arch_flags[arch])

    return command, version, features


class makedep_wrapper(action):
    """This is a wrapper around `cc -MM ...` to normalize the output and
    make it portable across compilers."""

    def __init__(self, cmd):
        action.__init__(self, cmd.name, self.makedep)
        self.cmd = cmd

    def subst(self, old, new):
        return self.cmd.subst(old, new)

    def map(self, fs):
        return self.cmd.map(fs)  # just forward variables from makedep

    def makedep(self, targets, sources):
        dfile = targets[0]._filename
        self.cmd(targets, sources)
        out = open(dfile).read()
        # tokenize the generated output...
        tokens = [token for line in out.split('\\\n') for token in line.split()]
        # ...and remove the first two tokens, e.g. '<file>.o:' and '<file>.c'
        headers = tokens[2:]
        # filter out system headers
        # TODO: make this configurable
        headers = [h for h in headers if not h.startswith('/usr/include')]
        # header paths are relative to the toplevel srcdir,
        # while we need them to be relative to the current module
        base = targets[0].module.srcdir
        relpath = lambda f, base: f if os.path.isabs(f) else os.path.relpath(f, base)
        headers = [relpath(h, base) + '\n' for h in headers]
        with open(dfile, 'w') as f:
            f.writelines(headers)


class makedep(action):

    command = 'gcc $(cppflags) -MM -o $(<) $(>)'
    cppflags = map(compiler.cppflags)
    cppflags += map(compiler.define, translate, prefix='-D')
    cppflags += map(compiler.include, translate, prefix='-I')


class compile(action):

    command = 'gcc $(cppflags) $(cflags) -c -o $(<) $(>)'
    cppflags = map(compiler.cppflags)
    cppflags += map(compiler.define, translate, prefix='-D')
    cppflags += map(compiler.include, translate, prefix='-I')
    cflags = map(compiler.cflags)
    cflags += map(compiler.link, select_if, 'shared', '-fPIC')


class link(action):

    command = 'gcc $(ldflags) -o $(<) $(>) $(libs)'
    ldflags = map(compiler.ldflags)
    ldflags += map(compiler.linkpath, translate, prefix='-L')
    ldflags += map(compiler.link, select_if, 'shared', '-shared')
    if platform.os == 'Darwin':
        ldflags += map(compiler.soname, translate, prefix='-Wl,-install_name -Wl,')
    else:
        ldflags += map(compiler.soname, translate, prefix='-Wl,-soname -Wl,')
    libs = map(compiler.libs, translate, prefix='-l')

    def submit(self, targets, sources):
        # sources may contain object files as well as libraries
        # Separate the two, and add the libraries to the libs variable.

        src, linkpath, libs = gcc.split_libs(sources)
        linkpath = [compiler.linkpath(l, base='') for l in linkpath]
        libs = [compiler.libs(l) for l in libs]
        fs = set(*libs + linkpath)
        for t in targets:
            t.features |= fs
        action.submit(self, targets, src)


class gcc(cc):

    makedep = makedep_wrapper(makedep())
    compile = compile()
    archive = action('ar rc $(<) $(>)')
    link = link()

    def __init__(self, name='gcc', command=None, version='', features=()):

        command, version, features = validate(self.__class__, command or 'gcc',
                                              version, features)
        cc.__init__(self, name=name, version=version)
        self.features |= features
        if command:
            # if command is of the form <prefix>-g++, make sure
            # to adjust the names of the other tools of the toolchain.
            prefix = command[:-3] if command.endswith('gcc') else ''
            self.makedep.subst('gcc', command)
            self.compile.subst('gcc', command)
            self.archive.subst('ar', prefix + 'ar')
            self.link.subst('gcc', command)

        irule(self.compile, types.obj, types.c)
        irule(self.archive, types.lib, types.obj)
        irule(self.link, types.bin, (types.obj, types.dso, types.lib))
        irule(self.link, types.dso, (types.obj, types.dso))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import, division, print_function

import platform as P
import re

os = P.system()
architecture = P.machine()


class platform(object):

    @property
    def os(self):
        return P.system()

    @property
    def architecture(self):
        return P.machine()


def split_triplet(value):
    """Split host triplet into (cpu, manufacturer, os) components."""

    # In the simplest cases the triplet uses the form <cpu>-<manufacturer>-<os>
    # However, the OS part may itself contain a hyphen (notably when the 'kernel'
    # is split into its own component), and the manufacturer may be left out,
    # leading to ambiguous names, e.g. <cpu>-<kernel>-<os>.
    # Try to detect such cases

    if re.match(r'(\w+)-(\w+)-linux-(\w+)', value):
        # re-merge the 'linux' kernel with the rest of the os spec.
        cpu, manufacturer, os = re.match(r'(\w+)-(\w+)-(.*)', value).groups()
    elif re.match(r'(\w+)-linux-(\w+)', value):
        manufacturer = 'unknown'
        # re-merge the 'linux' kernel with the rest of the os spec.
        cpu, os = re.match(r'(\w+)-(.*)', value).groups()
    else:
        cpu, manufacturer, os = re.match(r'(\w+)-(\w+)-(\w+)', value).groups()
    return cpu, manufacturer, os


if __name__ == '__main__':

    print('Platform: os={}, architecture={}'.format(os, architecture))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .. import types
from . import composite
from os.path import join, normpath


class object(composite):
    """Compile an object from one or more source files."""

    def __init__(self, *args, **kwds):
        composite.__init__(self, *args, type=types.obj, **kwds)

    @property
    def _filename(self):
        host = str(self.features.target.os) if 'target' in self.features else ''
        name = self.type.synthesize_name(self.name, host)
        return normpath(join(self.module.builddir, self.relpath, name))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from .utils import add_metaclass
from .feature import feature, set
from .action import action
from .error import ArgumentError
from collections import defaultdict
from copy import deepcopy
import logging

logger = logging.getLogger('tools')


class tool_type(type):
    def __init__(cls, name, bases, dict):
        """For any attribute of type 'action', set the action's name
        to the attribute name. Further, set the action's 'tool' attribute
        to be cls. This provides a means to look up virtual overloads. For
        example, if a rule refers to 'compiler.link', and the build is invoked
        with 'compiler=gcc', we can substitute 'gcc.link' for 'compiler.link'.
        """

        cls.tool = feature(name, feature('name', sub=True), feature('version', sub=True))

        for k, v in dict.items():
            if isinstance(v, action):
                v._cls = cls
                v.name = k

    def __getattribute__(cls, name):
        """Re-define action lookup.

        If 'A' defines 'action', a lookup via a subclass 'B' as 'B.action'
        needs to result in an action that is aware of it belonging to 'B'.

        To that end we make sure that 'B.action' and 'A.action' return
        distinct objects (each referring to the appropriate class 'A' and 'B'
        respectively)."""

        a = type.__getattribute__(cls, name)
        if isinstance(a, action) and a._cls is not cls:
            a = deepcopy(a)
            a._cls = cls
        return a


@add_metaclass(tool_type)
class tool(object):

    _instances = defaultdict(list)
    path_spec = ''

    @staticmethod
    def finish():
        tool._instances.clear()

    @classmethod
    def find_version_requirement(cls, features):
        """Find (and validate) any version requirements for 'cls' in 'features'."""
        f1 = None
        for c in cls.__mro__:
            if issubclass(c, tool) and c != tool:
                f2 = features[c.__name__] if (c.__name__ in features and
                                              features[c.__name__].version) else None
                if f2:
                    if f1 and f1.version != f2.version:
                        n1, n2 = f1._type.name, f2._type.name
                        raise ArgumentError('conflicting version requirements '
                                            '{}.version={} and {}.version={}'
                                            .format(n1, f1.version, n2, f2.version))
                    else:
                        f1 = f2
        return str(f1.version) if f1 and f1.version is not None else ''

    def __init__(self, name='', version='', features=()):
        name = name or self.__class__.__name__
        self.features = set.instantiate(features).copy()
        # Set built-in features
        for c in self.__class__.__mro__:
            if issubclass(c, tool) and c != tool:
                self.features += c.tool(name=name, version=version)

        # Clone all actions, so tool instances can fine-tune them individually.
        for a in dir(self):
            o = getattr(self, a)
            if isinstance(o, action):
                o = deepcopy(o)
                o._cls = self.__class__
                o._tool = self
                setattr(self, a, o)

        # Register the tool with all base classes
        for c in self.__class__.__mro__:
            if issubclass(c, tool) and c != tool:
                tool._instances[c].append(self)

        logger.info('instantiate {} (name={}, version={})'
                    .format(self.__class__.__name__, name, version))

    @property
    def name(self):
        return str(self.features[self.__class__.__name__].name)

    @property
    def version(self):
        return str(self.features[self.__class__.__name__].version)

    @property
    def id(self):
        v = self.version
        return '{}-{}'.format(self.name, v) if v else self.name

    @classmethod
    def instantiated(cls, fs=None):
        """Check whether a tool of the given type and feature-set
        is already instantiated."""
        if fs is None:
            return tool._instances[cls]
        else:
            fs = set.instantiate(fs)
            return any([t for t in tool._instances[cls] if t.features.matches(fs)])

    @classmethod
    def instances(cls, fs=None):
        """Return all known instances of cls that meet the feature requirements."""

        fs = set.instantiate(fs)
        try:  # Lookup failure is not an error
            cls.instance(fs)
        except Exception:
            pass
        return [t for t in tool._instances[cls] if t.features.matches(fs)]

    @classmethod
    def instance(cls, fs=None):
        """Find an instance of cls that meets the feature requirements."""

        fs = set.instantiate(fs)
        if not cls.instantiated(fs):
            try:
                cls(features=fs)
            except Exception as e:
                from . import debug
                logger.info('trying to instantiate {} yields "{}"'
                            .format(cls.__name__, e))
                if debug:
                    import traceback
                    traceback.print_exc()
        tools = [t for t in tool._instances[cls] if t.features.matches(fs)]
        if tools:
            return tools[0]
        else:
            raise ValueError('No matching tool "{}" found'.format(cls.__name__))


def info(name, fs=()):

    from importlib import import_module
    tools = []
    try:
        import faber.tools  # noqa F401
        mod = import_module('.{}'.format(name), 'faber.tools')
        tool = getattr(mod, name)
        tools += tool.instances(fs)
    except Exception:
        pass
    if not tools:
        msg = 'no tools of type {} found'.format(name)
        msg += ' matching {}.'.format(fs.essentials()) if fs else '.'
        print(msg)
        return
    table = [['tool', 'name', 'version']]
    for t in tools:
        table.append([t.__class__.__name__, t.name, t.version])
    width = map(max, zip(*[map(len, t) for t in table]))
    format = '{{:{}}}\t{{:{}}}\t{{:{}}}'.format(*width)
    for t in table:
        print(format.format(*t))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, notfile
from ..rule import depend
from .. import output
from .. import logging


def init(builddir):
    from .check import cache, logfiles, check
    check.cache = cache(builddir)
    check.logfiles = logfiles()


def finish():
    from .check import check
    if check.cache:
        check.cache.finish()
        check.cache = None
    check.logfiles.clear()


def reset(level):
    from .check import check
    check.logfiles.reset()
    if check.cache:
        check.cache.reset(level)


def clean(level):
    from .check import check
    if level > 1:
        check.logfiles.clean()
        if check.cache:
            check.cache.clean()


class report(artefact):

    def __init__(self, name, checks):
        use = [c.use for c in checks]
        artefact.__init__(self, name, attrs=notfile, use=use)
        depend(self, checks)
        self.checks = checks

    def _report(self):

        logger = logging.getLogger('summary')

        max_name_length = max(len(c.qname) for c in self.checks)
        logger.info(output.coloured('configuration check results:', attrs=['bold']))
        for c in self.checks:
            logger.info('  {:{}} : {} {}'
                        .format(c.qname, max_name_length, c.result, '(cached)' if c.cached else ''))

    def __status__(self, status):
        artefact.__status__(self, status)
        self._report()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..feature import set, map, translate, select_if
from .. import types
from .. import platform
from ..assembly import implicit_rule as irule
from . import compiler
from .cxx import *
from .clang import validate, makedep_wrapper


class makedep(action):

    command = 'clang++ $(cppflags) -MM -o $(<) $(>)'
    cppflags = map(compiler.cppflags)
    cppflags += map(compiler.define, translate, prefix='-D')
    cppflags += map(compiler.include, translate, prefix='-I')
    cppflags += map(cxxstd, translate, prefix='-std=c++')


class compile(action):

    command = 'clang++ $(cppflags) $(cxxflags) -c -o $(<) $(>)'
    cppflags = map(compiler.cppflags)
    cppflags += map(compiler.define, translate, prefix='-D')
    cppflags += map(compiler.include, translate, prefix='-I')
    cxxflags = map(compiler.cxxflags)
    cxxflags += map(cxxstd, translate, prefix='-std=c++')
    cxxflags += map(compiler.link, select_if, 'shared', '-fPIC')


class link(action):

    command = 'clang++ $(ldflags) -o $(<) $(>) $(libs)'
    ldflags = map(compiler.ldflags)
    ldflags += map(compiler.linkpath, translate, prefix='-L')
    ldflags += map(compiler.link, select_if, 'shared', '-shared')
    if platform.os == 'Darwin':
        ldflags += map(compiler.soname, translate, prefix='-Wl,-install_name -Wl,')
    else:
        ldflags += map(compiler.soname, translate, prefix='-Wl,-soname -Wl,')
    libs = map(compiler.libs, translate, prefix='-l')

    def submit(self, targets, sources):
        # sources may contain object files as well as libraries
        # Separate the two, and add the libraries to the libs variable.

        src, linkpath, libs = clangxx.split_libs(sources)
        linkpath = [compiler.linkpath(l, base='') for l in linkpath]
        libs = [compiler.libs(l) for l in libs]
        fs = set(*libs + linkpath)
        for t in targets:
            t.features |= fs
        action.submit(self, targets, src)


class clangxx(cxx):

    makedep = makedep_wrapper(makedep())
    compile = compile()
    archive = action('ar rc $(<) $(>)')
    link = link()

    def __init__(self, name='clang++', command=None, version='', features=()):

        command, version, features = validate(self.__class__, command or 'clang++',
                                              version, features)
        cxx.__init__(self, name=name, version=version)
        self.features |= features
        if command:
            self.makedep.subst('clang++', command)
            self.compile.subst('clang++', command)
            self.link.subst('clang++', command)

        irule(self.compile, types.obj, types.cxx)
        irule(self.archive, types.lib, types.obj)
        irule(self.link, types.bin, (types.obj, types.dso, types.lib))
        irule(self.link, types.dso, (types.obj, types.dso))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import operator
import re


def contains(op1, op2):
    # allow op1 to be `false`, so `set.nonexistent.contains('x')`
    # becomes a valid expression
    return False if isinstance(op1, false) else operator.contains(op1, op2)


def matches(op1, op2):
    # allow op1 to be `false`, so `set.nonexistent.match('x')`
    # becomes a valid expression
    return False if isinstance(op1, false) else re.match(op2, str(op1))


class expr(object):

    # we can't overload __not__ as that has to return a bool...
    def not_(self): return unary(operator.not_, self)
    def __eq__(self, other): return binary(operator.eq, self, other)
    def __ne__(self, other): return binary(operator.ne, self, other)
    def __gt__(self, other): return binary(operator.gt, self, other)
    def __lt__(self, other): return binary(operator.lt, self, other)
    def __and__(self, other): return binary(operator.and_, self, other)
    def __or__(self, other): return binary(operator.or_, self, other)
    # we can't overload __contains__ as that has to return a bool...
    def contains(self, other): return binary(contains, self, other)
    def __bool__(self): raise ValueError('invalid expression "{}" !'.format(self))
    def __nonzero__(self): return self.__bool__()
    def __call__(self, ctx): return True
    def matches(self, other): return binary(matches, self, other)


class true(expr):

    def __eq__(self, other): return True if bool(other) else False
    def __bool__(self): return True
    def __call__(self, ctx): return True
    def __str__(self): return '<expr True>'


class false(expr):

    def __eq__(self, other): return True if not other else False
    def __bool__(self): return False
    def __call__(self, ctx): return False
    def __str__(self): return '<expr False>'


class unary(expr):

    def __init__(self, op, op1):
        self.op = op
        self.op1 = op1

    def __call__(self, ctx):
        op1 = self.op1(ctx)
        return self.op(op1)

    def __str__(self):
        return '<expr {}({})>'.format(self.op.__name__, self.op1)


class binary(expr):

    def __init__(self, op, op1, op2):
        self.op = op
        self.op1 = op1
        self.op2 = op2

    def __call__(self, ctx):
        op1 = self.op1(ctx) if isinstance(self.op1, expr) else self.op1
        op2 = self.op2(ctx) if isinstance(self.op2, expr) else self.op2
        return self.op(op1, op2)

    def __str__(self):
        return '<expr {}({}, {})>'.format(self.op.__name__, self.op1, self.op2)


class sub(expr):
    # a subfeature
    def __init__(self, op, name):
        self.op = op
        self.name = name

    def __call__(self, ctx):
        op = self.op(ctx)
        return getattr(op, self.name) if hasattr(op, self.name) else false()

    def __str__(self):
        return '<expr attr({}, {})>'.format(self.op, self.name)


class value(expr):
    def __init__(self, name):
        self._name = name

    def __getattr__(self, name):
        return sub(self, name)

    def __call__(self, ctx):
        return ctx[self._name] if self._name in ctx else false()

    def __str__(self):
        return '<expr value({})>'.format(self._name)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from . import termcolor
import sys

if sys.platform == 'win32':
    try:
        from colorama import init
        init()
        coloured = termcolor.colored
    except Exception:
        # fall back to undecorated text
        coloured = lambda text, colour=None, on_colour=None, attrs=None: text
else:
    coloured = termcolor.colored

if __name__ == '__main__':

    print(coloured('this is red', 'red'))
    print(coloured('this is green', 'green'))
    print(coloured('this is bold', attrs=['bold']))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .check import check
from ..artefact import intermediate, always
from ..tools.compiler import compiler
from ..rule import rule
from ..action import action
from ..artefacts.binary import binary
import subprocess


class try_run(check):
    """Try to compile and run a chunk of source code."""

    run = action('run', '$(>)')

    def __init__(self, name, source, type, features=(), if_=(), ifnot=()):

        check.__init__(self, name, features, if_, ifnot)
        compiler.check_instance_for_type(type, self.features)
        if not self.cached:
            # create source file
            src = type.synthesize_name(self.name)

            def generate(targets, _):
                with open(targets[0]._filename, 'w') as os:
                    os.write(source)
            src = rule(generate, src, attrs=intermediate|always,
                       logfile=self.logfile)
            bin = binary(self.name, src, features=self.features,
                         attrs=intermediate, logfile=self.logfile)
            rule(self.run, self, bin, logfile=self.logfile)


class check_output(try_run):
    """Compile and run a chunk of source code and check the generated output."""

    def post_process(self, output):
        self.result = output

    def run(self, _, source):
        """run the binary in a subprocess, then post-process the output."""
        output = subprocess.check_output([source[0].filename.result()]).decode().strip()
        self.post_process(output)



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils import nodes
from docutils.statemachine import ViewList
from sphinx.domains import Domain, ObjType, Index
from sphinx.directives import ObjectDescription
from sphinx.roles import XRefRole
from sphinx import addnodes
from sphinx.ext import graphviz


class Field(object):

    def __init__(self, name, label):
        self.name = name
        self.label = label
    
    def make_field(self, directive):
        # treat content like docfields.Field
        value = directive.options.get(self.name)
        if value:
            name = nodes.field_name('', self.label)
            para = nodes.paragraph(rawsource=value)
            content = ViewList()
            content.append(value, *directive.content.items[0])
            directive.state.nested_parse(content, directive.content_offset, para)
            body = nodes.field_body('', para)
            return nodes.field('', name, body)


class XRefField(Field):

    def __init__(self, name, label, domain):
        super(XRefField, self).__init__(name, label)
        self.domain = domain
    
    def make_xref(self, rolename, target):
        refnode = addnodes.pending_xref('',
                                        refdomain=self.domain,
                                        reftype=rolename,
                                        reftarget=target)
        refnode += addnodes.literal_emphasis(target, target)
        return refnode

    def make_field(self, directive):
        # treat content like docfields.Field
        value = directive.options.get(self.name)
        if value:
            name = nodes.field_name('', self.label)
            para = nodes.paragraph(rawsource=value)
            value = value.split(',')
            refs = [self.make_xref('', v.strip()) for v in value]
            content = [nodes.Text(',')]*(2*len(refs)-1)
            content[::2] = refs
            para += content
            body = nodes.field_body('', para)
            return nodes.field('', name, body)


class InheritanceField(Field):

    def make_field(self, directive):
        # treat content like docfields.Field
        value = directive.options.get(self.name)
        if value:
            name = nodes.field_name('', self.label)
            para = nodes.paragraph(rawsource=value)
            bases = value.split(',')
            # graph attributes
            code = ['rankdir=LR;']
            code += ['size="8.0, 12.0";']
            t = directive.arguments[0]
            # node attributes
            node_attrs = ', '.join(['shape="box"',
                                    'fontsize="10"',
                                    'height="0.25"',
                                    'fontname="Vera Sans, DejaVu Sans, Liberation Sans, Arial, Helvetica, sans"',
                                    'style="setlinewidth(0.5) filled"',
                                    'fillcolor="khaki3:khaki1"'])
            code += ['%s [%s, label="%s"];'%(t, node_attrs, t)]
            code += ['%s [%s, label="%s"];'%(b, node_attrs, b) for b in bases]
            code += ['%s -> %s;'%(t, b) for b in bases]
            node = graphviz.graphviz()
            node['code'] = 'digraph I {\n%s\n}\n'%'\n'.join(code)
            node['options'] = {}
            body = nodes.field_body('', node)        
            return nodes.field('', name, body)


class FaberDirective(ObjectDescription):
    """Inject field options into content."""

    has_content = True
    required_arguments = 1
    optional_arguments = 0
    fields = []

    def run(self):
        index, content = ObjectDescription.run(self)
        fields = nodes.field_list()
        for f in self.fields:
            field = f.make_field(self)
            if field:
                fields += field
        # Insert option fields right after signature
        content.insert(1, fields)

        name = self.arguments[0]
        module = self.options.get('module', None)
        qname = '{}.{}'.format(module, name) if module else name
        classname = self.name[4:] if self.name.startswith('fab:') else self.name
        indextext = '{} ({} in {})'.format(name, classname, module)
        index = addnodes.index(entries=[('single', indextext, name, None, None)])
        return [index, content]


class Feature(FaberDirective):
    
    option_spec = dict(module=directives.unchanged,
                       attributes=directives.unchanged,
                       values=directives.unchanged)
    fields = [Field('module', 'Defined in'),
              XRefField('attributes', 'Attributes', 'py'),
              Field('values', 'Possible values')]

    
class Action(FaberDirective):

    option_spec = dict(features=directives.unchanged,
                       module=directives.unchanged,
                       abstract=directives.unchanged)
    fields = [Field('features', 'Features used')]


class Tool(FaberDirective):
    
    option_spec = dict(module=directives.unchanged,
                       bases=directives.unchanged,
                       actions=directives.unchanged)
    fields = [Field('module', 'Defined in'),
              InheritanceField('bases', 'Inheritance'),
              Field('actions', 'Actions')]


class FaberDomain(Domain):
    """Faber domain."""

    name = 'fab'
    label = 'Faber'

    object_types = dict(feature=ObjType('feature', 'feature', 'obj'),
                        tool=ObjType('tool', 'tool', 'obj'),
                        action=ObjType('action', 'action', 'obj'))

    directives = dict(feature=Feature,
                      tool=Tool,
                      action=Action)

    roles = dict(feature=XRefRole(),
                 tool=XRefRole(),
                 action=XRefRole())

    initial_data = dict(feature={})  # path: (docname, synopsis)

    #indices = [FeatureIndex, ToolIndex]


def setup(app):

    app.add_domain(FaberDomain)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import asyncio
from enum import Enum, Flag, IntEnum
from os.path import dirname, lexists
from os import stat, makedirs, remove, rmdir
from collections import defaultdict
import logging

logger = logging.getLogger('scheduler')
summary_logger = logging.getLogger('summary')

flag = Flag('flag',
            'NONE '
            'TEMP NOCARE NOTFILE TOUCHED '
            'LEAVES NOUPDATE XX RMOLD '
            'INTERNAL PRECIOUS NOPROPAGATE', start=0)
binding = Enum('binding', 'UNBOUND MISSING PARENTS EXISTS')
fate = IntEnum('fate',
               'INIT MAKING STABLE NEWER SPOIL '
               'ISTMP BUILD TOUCHED REBUILD MISSING NEEDTMP '
               'OUTDATED UPDATE BROKEN CANTFIND CANTMAKE')
progress = IntEnum('progress',
                   'INIT LAUNCHED BOUND RUNNING DONE NOEXEC_DONE')


def _cyclic(artefact, root=None):
    if root == artefact:
        yield True
    root=root or artefact
    for p in artefact.prerequisites:
        yield from _cyclic(p, root)


def cyclic(artefact):
    return next(_cyclic(artefact), False)


class dependency_error(Exception): pass


class artefact(object):

    @classmethod
    def init(cls, files=[], keep_temps=False, force=False):
        """set up some global state."""
        cls.counter = defaultdict(int)
        cls.files = files
        cls.temp_files = set()
        cls.keep_temps = keep_temps
        cls.force = force

    @classmethod
    def finish(cls):
        """clean up."""
        # Remove temp files...
        if not cls.keep_temps:
            for i in cls.temp_files:
                if lexists(i):
                    remove(i)
        # ...and print a cumulative summary
        cls._format_count('...failed updating {} artefact{}...', 'failed')
        cls._format_count('...skipped {} artefact{}...', 'skipped')
        cls._format_count('...updated {} artefact{}...', 'updated')

        del cls.files

    @classmethod
    def clean(cls, level):
        """Remove generated files."""
        dirs = []
        # remove file artefacts...
        for f in cls.files:
            # We may encounter symbolic links during the cleanup
            # which no longer refer to existing files. To be able
            # to detect them we need to use `lexists`...
            if lexists(f):
                dirs.append(dirname(f))
                remove(f)
        if cls.files:
            cls.files.clear()
        # ...then clean up empty build directories
        dirs = sorted(set(dirs), reverse=True)
        root = '' if cls.files.root == '.' else cls.files.root
        for d in dirs:
            try:
                while d != root:
                    rmdir(d)
                    d = dirname(d)
            except OSError:  # directory wasn't empty, move on
                continue

    @classmethod
    def _format_count(cls, msg, what):
        number = cls.counter[what]
        if number:
            summary_logger.info(msg.format(number, 's' if number > 1 else ''))

    def __init__(self, frontend, prerequisites=[]):
        self.frontend = frontend
        self.boundname = None
        self.recipe = None
        self.prerequisites = set(prerequisites)
        self._lock = asyncio.Lock()
        self._pqueue = None
        self._dependants = set()
        self._rebuilds = set()    # targets that should be force-rebuilt whenever this one is
        self.flags = flag(self.frontend.attrs)
        self.reset()

    def reset(self):
        self.binding = binding.UNBOUND
        self._timestamp = 0
        self._fate = fate.INIT
        self.progress = progress.INIT
        self.status = None

    @property
    def isfile(self):
        return not self.flags & flag.NOTFILE

    @property
    def name(self):
        return self.frontend.name

    @property
    def timestamp(self):
        """Report the artefact's modification time (0 for non-file artefacts).
        For temporaries report the real timestamp only if the file exists.
        Otherwise return the most recent time among its prerequisites."""

        if self.binding == binding.EXISTS:
            return self._timestamp
        elif self.flags & flag.TEMP:  # report most recent prerequisite timestamps
            return max([p.timestamp for p in self.prerequisites], default=0)
        else:
            return 0

    @property
    def fate(self):
        """Return the artefact's fate. For temporaries report the least stable fate of its prerequisites."""
        return max([p.fate for p in self.prerequisites], default=fate.INIT) if self.flags & flag.TEMP else self._fate

    def add_prerequisite(self, p):
        """Add p to the artefact's set of prerequisites.
        This may fail for different reasons:
        * the prerequisite set is immutable once the artefact is bound.
        * the new prerequisite must not introduce a dependency cycle."""

        if self.progress >= progress.BOUND:
            raise dependency_error(f'can not add {p.frontend}: '
                                   f'{self.frontend.boundname} already bound')
        self.prerequisites.add(p)
        if cyclic(self):
            raise dependency_error(f'dependency cycle detected while adding '
                                   f'{self.frontend} -> {p.frontend}')
        if self._pqueue:
            self._pqueue.put_nowait(p)

    async def process(self, parent=None):
        """Process this artefact:
        * process all prerequisites
        * bind it
        * determine its fate
        * update it"""

        await self.launch(parent)
        # at this point all prereqs are bound, non-temps will even be updated.
        await self.bind(parent)
        # temps require parents to be bound first, to determine their fate
        if not self.flags & flag.TEMP or not parent:
            await self.compute_fate(parent)
            await self.update()

    async def launch(self, parent=None):
        """Update all prerequisites."""
        async with self._lock:
            if self.progress >= progress.LAUNCHED:
                return
            logger.info(f'progress -- {self.frontend} launching')
            # the prerequisite set may grow while we process it, so we use a queue.
            self._pqueue = asyncio.Queue()
            await asyncio.gather(*[self._pqueue.put(p) for p in self.prerequisites])
            while not self._pqueue.empty():
                p = await self._pqueue.get()
                await p.process(self)
            self._pqueue = None
            self.progress = progress.LAUNCHED
            logger.info(f'progress -- {self.frontend} launched')

    async def bind(self, parent=None):
        """Bind all artefact-specific variables, including the artefact's filename.
        Precondition: the prerequisite set is now immutable, and all prerequisites are bound."""

        async with self._lock:
            if self.progress >= progress.BOUND:
                return
            try:
                self.frontend.features.eval(update=False)
            except Exception as e:
                logger.critical(f'something went wrong binding {self.frontend}: {e}')
                raise
            self.boundname = self.frontend.boundname
            if not self.flags & flag.NOTFILE:
                d = dirname(self.boundname) or '.'
                if not lexists(d):
                    makedirs(d)
                self.binding = binding.EXISTS if lexists(self.boundname) else binding.MISSING
                self._timestamp = stat(self.boundname).st_mtime if self.binding == binding.EXISTS else 0

            # if temp file does not exist but parent does, use parent
            if (parent and
                self.flags & flag.TEMP and
                self.binding == binding.MISSING and
                parent.binding != binding.MISSING):
                self.binding = binding.PARENTS

            msg = f'bind -- {id(self)} {self.name}: {self.boundname} '
            msg += f'time={self._timestamp}' if self.binding == binding.EXISTS else f'{str(self.binding)}'
            logger.info(msg)
            self.progress = progress.BOUND
            logger.info(f'progress -- {self.frontend} bound')

    async def compute_fate(self, parent=None):

        async with self._lock:
            if self._fate != fate.INIT:
                return
            self._fate=fate.MAKING

            for p in self.prerequisites:
                await p.compute_fate(self)

            self._fate = fate.STABLE
            last = 0
            for p in self.prerequisites:
                if p.flags & flag.NOPROPAGATE:
                    continue
                last = max(last, p.timestamp)
                if self._fate < p.fate:
                    logger.info(f'fate -- change {self.boundname} from {str(self._fate)} to {str(p.fate)} by dependency')
                    self._fate = p.fate
            else:
                # if this a (non-existing) temporary without prerequisites, treat it MISSING
                if self.flags & flag.TEMP:
                    self._fate = fate.MISSING
            if self.flags & flag.NOUPDATE:
                logger.info(f'fate -- change {self.boundname} back to stable, NOUPDATE')
                self._fate = fate.STABLE
            # If can not find or make child, can not make target.
            elif self._fate >= fate.BROKEN:
                self._fate = fate.CANTMAKE
            # If children changed, make target.
            elif self._fate >= fate.SPOIL:
                self._fate = fate.UPDATE
            # If target missing, make it.
            elif self.binding == binding.MISSING:
                self._fate = fate.MISSING
            # If children newer, make target.
            elif self.binding == binding.EXISTS and last > self.timestamp:
                self._fate = fate.OUTDATED
            # If temp's children newer than parent, make temp.
            elif self.binding == binding.PARENTS and last > parent.timestamp:
                self._fate = fate.NEEDTMP
            # If deliberately touched, make it.
            elif self.flags & flag.TOUCHED:
                self._fate = fate.TOUCHED
            # If force flag is set, make it.
            elif artefact.force:
                self._fate = fate.TOUCHED
            # If up-to-date temp file present, use it.
            # If target newer than non-notfile parent, mark target newer.
            # Otherwise, stable!

            if self._fate == fate.MISSING and not self.recipe and not self.prerequisites:
                if self.flags & flag.NOCARE:
                    self._fate = fate.STABLE
                else:
                    self._fate = fate.CANTFIND
            logger.info(f'fate -- {self.boundname}: {str(self._fate)}')

    async def update(self):

        async with self._lock:
            if self.progress == progress.DONE:
                return

            if self._fate != fate.STABLE:
                # Update temporary prerequisites
                await asyncio.gather(*[p.update() for p in self.prerequisites if p.flags & flag.TEMP and p._fate != fate.STABLE])

            failed = None
            if self._fate != fate.STABLE:
                failed = next(iter([p for p in self.prerequisites
                                    if p.fate != fate.STABLE and not p.status and not p.flags & flag.NOCARE]), None)

            # if a prerequisite failed, fail.
            if failed:
                self.status = False
            elif self._fate in (fate.STABLE, fate.NEWER):
                self.status = True
            elif self._fate >= fate.CANTFIND:
                self.status = False
            elif self._fate == fate.ISTMP:
                summary_logger(f'...using {self.boundname}...')
                self.status = True
            elif self._fate >= fate.TOUCHED:
                if self.recipe:
                    self.progress = progress.RUNNING
                    logger.info(f'progress -- {self.frontend} running')
                    logger.info(f'update -- {self.boundname}')
                    self.status = await self.recipe()
                    if self.flags & flag.TEMP:
                        artefact.temp_files.add(self.boundname)
                    elif not self.flags & flag.NOTFILE:
                        artefact.files.append(self.boundname)
                    logger.info(f'update -- {self.boundname} done (status={self.status})')
                    artefact.counter['updated' if self.status else 'failed'] += 1
                else:
                    # TODO: how should we handle alias artefacts ? What if prereqs fail ? Etc.
                    self.status = True
            self.progress = progress.DONE
            logger.info(f'progress -- {self.frontend} done')
            self._report(failed)

    def _report(self, failed):

        if failed:
            msg = f'...skipped {self.boundname} for lack of {failed.boundname}...'
            if self.frontend.logfile:
                self.frontend.logfile.write(msg + '\n')
            else:
                summary_logger.info(msg)
        self.frontend.__status__(self.status)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import artefact, intermediate
from faber import assembly
from faber.action import action
from faber.types import *
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_implicit_rule():

    a1 = action('a1', 'something')
    a2 = action('a2', 'something')
    a3 = action('a3', 'something')
    c_ = artefact('c', type=c)
    cc = artefact('cc', type=cxx)
    o = artefact('o', type=obj)
    l = artefact('l', type=lib)
    assembly.implicit_rule(a1, obj, c)
    assembly.implicit_rule(a2, obj, cxx)
    assembly.implicit_rule(a3, lib, obj)
    with patch('faber.assembly.explicit_rule') as rule:
        assembly.rule(o, c_)
        assembly.rule(o, cc)
        assert rule.call_count == 2
        assert (rule.call_args_list[0][0][0] is a1 and
                rule.call_args_list[0][0][1] is o)
        assert (rule.call_args_list[1][0][0] is a2 and
                rule.call_args_list[1][0][1] is o)

    with patch('faber.assembly.explicit_rule') as rule:
        assembly.rule(l, cc)
        assert rule.call_count == 2
        assert (rule.call_args_list[0][0][0] is a2 and
                rule.call_args_list[0][1]['attrs'] & intermediate)
        assert (rule.call_args_list[1][0][0] is a3 and
                rule.call_args_list[1][0][1] is l)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from .artefact import artefact
from .artefact import dependency_error as DependencyError  # noqa F401
from .recipe import recipe
from ..cache import filecache
from ..utils import aslist
import asyncio
import sys

__all__ = ['init', 'reset', 'clean', 'finish',
           'variables', 'define_artefact', 'add_dependency', 'define_recipe',
           'run', 'update', 'print_dependency_graph', 'DependencyError']

if sys.platform == 'win32':
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


artefacts = {}  # map frontends to backends


def init(params, builddir, readonly=False, **options):
    noexec = options.get('noexec', False)
    files = filecache(builddir, params) if not readonly else ()
    intermediates = options.get('intermediates', False)
    jobs = options.get('parallel', 1)
    timeout = options.get('timeout', 0)
    force = options.get('force', False)
    artefact.init(files, intermediates, force)
    recipe.init(jobs, timeout, noexec)


def reset():
    for b, f in artefacts.items():
        b.reset()
        f.reset()


def clean(level=1):
    artefact.clean(level)


def finish():
    artefact.finish()


def variables(a):
    recipe = artefacts[a].recipe
    return recipe.variables() if recipe else {}


def define_artefact(a, bind=False):
    artefacts[a] = artefact(a)


def add_dependency(a, deps):
    for d in [artefacts[d] for d in aslist(deps)]:
        artefacts[a].add_prerequisite(d)


def define_recipe(a, targets, sources=[]):
    targets = [artefacts[t] for t in targets]
    sources = [artefacts[s] for s in sources]
    r = recipe(action=a, targets=targets, sources=sources)
    for t in targets:
        t.recipe = r


def run(command):
    return recipe.run_subprocess(command)


def update(aa):
    try:
        loop = asyncio.get_event_loop()
        aa = [artefacts[a] for a in aslist(aa)]
        loop.run_until_complete(asyncio.gather(*[a.process() for a in aa]))
        return all([a.status for a in aa])
    except Exception:
        raise


def print_dependency_graph(aa=[]):
    from . import graph
    graph.visualize(*[artefacts[a] for a in aslist(aa)], filename='dependencies.png')



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)


class InvalidState(Exception):

    def __init__(self, a):
        cln = a.__module__ + '.' + a.__class__.__qualname__
        Exception.__init__(self, f'cannot access unfinished {cln} {a}')
        self.artefact = a


class delayed(object):
    """A delayed value may only be used after a prerequisite artefact has
    been updated."""

    def __init__(self, f, a):
        self._func = f if callable(f) else lambda: f  # the wrapped value
        self._artefact = a  # the artefact on which the value depends

    def apply(self, f):
        """Create a new delayed object representing the value `f(self.result())`"""
        return delayed(lambda: f(self._func()), self._artefact)

    def result(self):
        if self._artefact.status is None:
            raise InvalidState(self._artefact)
        return self._func()

    def __str__(self):
        cln = self.__class__.__name__
        return f'<{cln} func={self._func} artefact={self._artefact}>'


class delayed_property(property):

    def __init__(*args, **kwds):
        property.__init__(*args, **kwds)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return delayed(lambda: self.fget(instance), instance)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..feature import set
from ..tool import tool
from .xslt import process
import os
from os.path import join


class quickbook(tool):

    process = action('quickbook --input-file=$(>) --output-file=$(<)')

    def __init__(self, name='quickbook', command=None, version='', features=()):
        tool.__init__(self, name=name, version=version)
        self.features |= features
        if command:
            self.process.subst('quickbook', command)


class boostbook(tool):

    db = process()  # bb -> db
    html = process()  # db -> html

    def __init__(self, name='boostbook', command=None, version='', prefix='', features=()):
        tool.__init__(self, name=name, version=version)
        self.features |= set.instantiate(features)
        if command:
            self.db.subst('xsltproc', command)
            self.html.subst('xsltproc', command)
        if not prefix and 'BOOST_ROOT' in os.environ:
            prefix = join(os.environ['BOOST_ROOT'], 'tools/boostbook')
        if prefix:
            db_xsl = prefix + '/xsl/docbook.xsl'
            self.db.subst('$(stylesheet)', db_xsl)
            html_xsl = prefix + '/xsl/html.xsl'
            self.html.subst('$(stylesheet)', html_xsl)
        else:
            raise ValueError('no prefix set and $BOOST_ROOT undefined')



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber import cli
from test.common import cwd, argv
import pytest
import sys
from os.path import join

compilers = {'gcc': {'cc': 'gcc', 'cxx': 'g++'},
             'clang': {'cc': 'clang', 'cxx': 'clang++'},
             'msvc': {'cc': 'msvc', 'cxx': 'msvc'}}
if sys.platform=='win32':
    compilers['native'] = compilers['msvc']
else:
    compilers['native'] = compilers['gcc']


def get_cc_opt(name):
    return 'cc.name={}'.format(compilers[name]['cc'])


def get_cxx_opt(name):
    return 'cxx.name={}'.format(compilers[name]['cxx'])


def faber(*args):
    with argv(('faber',) + args):
        return cli.main()


if sys.platform=='win32':
    # on Windows we can't clean up the config cache while
    # it is still being accessed via the test module itself
    clean = '-c'  # clean
else:
    clean = '-cc'  # extra clean


@pytest.mark.skipif(sys.platform == 'win32', reason='requires a "c++" tool')
def test_action():

    with cwd(join('examples', 'action')):
        assert faber()
        assert faber(clean)


def test_tool(compiler):

    args = []
    if compiler:
        args.append(get_cxx_opt(compiler))
    with cwd(join('examples', 'tool')):
        assert faber(*args)
        assert faber(clean)


def test_implicit_rules(compiler):

    args = []
    if compiler:
        args.append(get_cxx_opt(compiler))
    with cwd(join('examples', 'implicit_rules')):
        assert faber(*args)
        assert faber(clean)


def test_modular(compiler):

    args = []
    if compiler:
        args.append(get_cxx_opt(compiler))
    with cwd(join('examples', 'modular')):
        assert faber(*args)
        assert faber(clean)


def test_config(compiler):

    args = []
    if compiler:
        args.append(get_cxx_opt(compiler))
    with cwd(join('examples', 'config')):
        assert faber(*args)
        assert faber(clean)


def test_python(compiler):

    args = []
    if compiler:
        args.append(get_cc_opt(compiler))
    with cwd(join('examples', 'python')):
        assert faber(*args)
        assert faber(clean)


def test_test(compiler):

    args = []
    if compiler:
        args.append(get_cxx_opt(compiler))
    with cwd(join('examples', 'test')):
        assert faber(*args)
        assert faber(clean)


def test_package(compiler):

    args = []
    if compiler:
        args.append(get_cxx_opt(compiler))
    with cwd(join('examples', 'package')):
        assert faber(*args + ['tgz', 'stgz'])
        assert faber(clean)



#
# Copyright (c) 2019 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .test import *
from ..artefact import artefact, notfile


class suite(artefact):
    """A test suite provides an alias to a set of tests."""

    def __init__(self, name, tests, attrs=notfile, module=None):
        """tests consists of either test objects or other suites."""
        super(suite, self).__init__(name, attrs=attrs|notfile, module=module)
        self.tests = tests
        depend(self, self.tests)

    def __iter__(self):
        for t in self.tests:
            if isinstance(t, suite):
                # Use `yield from` once we stop supporting Python 2.7
                for i in t:
                    yield i
            else:
                yield t



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .. import types
from .. import assembly
from . import composite
from ..artefact import artefact, delayed_property
from ..action import action
from ..rule import rule
from .. import platform
from ..tools.installer import installer
from .install import _installed
from os.path import dirname, basename, normpath, join


class library(composite):

    # This command assumes that source an target live in the same directory.
    symlink = action('symlink', 'cd `dirname $(>)` && ln -s `basename $(>)` `basename $(<)`')

    def __init__(self, *args, **kwds):
        version = kwds.pop('version', None)
        self.version = [str(v) for v in version] if version else None
        self._versioned = None  # to be set in _assemble
        composite.__init__(self, *args, type=types.lib, **kwds)

    def _assemble(self):
        # make sure all compiler features are instantiated
        from ..tools.compiler import compiler, link, soname  # noqa F401
        self.features.eval(update=False)
        if 'link' not in self.features:
            self.features += link('shared')
        self.type = types.dso if self.features.link == 'shared' else types.lib
        if (self.version and self.features.link == 'shared'):
            target = str(self.features.target.os) if 'target' in self.features else platform.os
            if target not in ('Windows'):
                base = basename(self._filename)
                # a versioned shared library allows multiple versions to coexist.
                # set SONAME if there is a bugfix version included
                if len(self.version) == 3:
                    self.features += soname('{}.{}.{}'.format(base, self.version[0], self.version[1]))
                a = artefact(base + '.' + '.'.join(self.version),
                             type=self.type, features=self.features, path_spec=self.path_spec, module=self.module)
                self._versioned = assembly.rule(a, self.sources, features=self.features, module=self.module)
                suffixes = list(reversed(['.'.join(self.version[:i]) for i in range(len(self.version))]))
                links = [join(a.relpath, '{}.{}'.format(base, suffixes[i])) for i in range(len(suffixes) - 1)]
                self.path_spec = a.path_spec
                self.features = a.features
                links.append(self)
                for l in links:
                    a = rule(library.symlink, l, a)
                return
        assembly.rule(self, self.sources, self.features, module=self.module)

    @property
    def libname(self):
        return basename(self.name)

    @property
    def _filename(self):
        dir, base = dirname(self.name), self.libname
        host = str(self.features.target.os) if 'target' in self.features else ''
        libname = join(dir, self.type.synthesize_name(base, host))
        return normpath(join(self.module.builddir, self.relpath, libname))

    @delayed_property
    def path(self):
        return dirname(self._filename)


class installed(_installed):
    # installing a library may require creating symbolic links, if the library is versioned

    @property
    def _filename(self):
        # use the name of the upstream artefact, combined with our relpath
        dir, base = dirname(self.a.name), self.a.libname
        host = str(self.a.features.target.os) if 'target' in self.a.features else ''
        libname = join(dir, self.a.type.synthesize_name(base, host))
        return normpath(join(self.relpath, libname))

    def _define_rule(self):
        if self.a._versioned:
            v = _installed(self.a._versioned, self.subdir, self.features)
            base = basename(self._filename)
            suffixes = list(reversed(['.'.join(self.a.version[:i]) for i in range(len(self.a.version))]))
            links = [join(self.relpath, '{}.{}'.format(base, suffixes[i])) for i in range(len(suffixes) - 1)]
            for l in links:
                v = rule(library.symlink, l, v)
            rule(library.symlink, self, v)
        else:
            rule(installer.install, self, self.a)



#
# Copyright (c) 2020 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import source, artefact


def make_source(name, content):
    """Create and return a source artefact with the specified content."""
    src = source(name)
    with open(src._filename, 'w') as out:
        out.write(content)
    return src


def check_makedep(compiler, src, num_deps):
    """Generic makedep test. Call with a specific compiler instance."""

    dependencies = artefact('dependencies.d')
    targets, sources = [dependencies], [src]
    compiler.makedep(targets, sources)
    with open(dependencies._filename, 'r') as d:
        assert sum(1 for line in d if line) == num_deps


def check_compile(compiler, src):
    """Generic compile test. Call with a specific compiler instance."""

    obj = artefact('out.o')
    targets, sources = [obj], [src]
    compiler.compile(targets, sources)
